from __future__ import annotations

import copy
import os

import pytest
import ray

import daft
from daft import udf
from daft.daft import SystemInfo
from daft.expressions import col
from daft.internal.gpu import cuda_visible_devices
from tests.conftest import get_tests_daft_runner_name


def no_gpu_available() -> bool:
    return len(cuda_visible_devices()) == 0


DATA = {"id": [i for i in range(100)]}


@udf(return_dtype=daft.DataType.int64())
def my_udf(c):
    return [1] * len(c)


###
# Test behavior of overriding options
###


def test_partial_resource_request_overrides():
    new_udf = my_udf.override_options(num_cpus=1.0)
    assert new_udf.resource_request.num_cpus == 1.0
    assert new_udf.resource_request.num_gpus is None
    assert new_udf.resource_request.memory_bytes is None

    new_udf = new_udf.override_options(num_gpus=8.0)
    assert new_udf.resource_request.num_cpus == 1.0
    assert new_udf.resource_request.num_gpus == 8.0
    assert new_udf.resource_request.memory_bytes is None

    new_udf = new_udf.override_options(num_gpus=None)
    assert new_udf.resource_request.num_cpus == 1.0
    assert new_udf.resource_request.num_gpus is None
    assert new_udf.resource_request.memory_bytes is None

    new_udf = new_udf.override_options(memory_bytes=100)
    assert new_udf.resource_request.num_cpus == 1.0
    assert new_udf.resource_request.num_gpus is None
    assert new_udf.resource_request.memory_bytes == 100


def test_resource_request_pickle_roundtrip():
    new_udf = my_udf.override_options(num_cpus=1.0)
    assert new_udf.resource_request.num_cpus == 1.0
    assert new_udf.resource_request.num_gpus is None
    assert new_udf.resource_request.memory_bytes is None

    assert new_udf == copy.deepcopy(new_udf)

    new_udf = new_udf.override_options(num_gpus=8.0)
    assert new_udf.resource_request.num_cpus == 1.0
    assert new_udf.resource_request.num_gpus == 8.0
    assert new_udf.resource_request.memory_bytes is None
    assert new_udf == copy.deepcopy(new_udf)


@pytest.mark.skipif(get_tests_daft_runner_name() != "native", reason="requires Native Runner to be in use")
def test_requesting_too_many_cpus():
    df = daft.from_pydict(DATA)

    my_udf_parametrized = my_udf.override_options(num_cpus=1000)
    df = df.with_column(
        "foo",
        my_udf_parametrized(col("id")),
    )

    with pytest.raises(Exception):
        df.collect()


@pytest.mark.skipif(get_tests_daft_runner_name() != "native", reason="requires Native Runner to be in use")
def test_requesting_too_much_memory():
    df = daft.from_pydict(DATA)
    system_info = SystemInfo()

    my_udf_parametrized = my_udf.override_options(memory_bytes=system_info.total_memory() + 1)
    df = df.with_column(
        "foo",
        my_udf_parametrized(col("id")),
    )

    with pytest.raises(Exception):
        df.collect()


###
# Assert RayRunner behavior for requests
###


@udf(return_dtype=daft.DataType.int64())
def assert_resources(c, num_cpus=None, num_gpus=None, memory=None):
    assigned_resources = ray.get_runtime_context().get_assigned_resources()

    for resource, ray_resource_key in [(num_cpus, "CPU"), (num_gpus, "GPU"), (memory, "memory")]:
        if resource is None:
            assert ray_resource_key not in assigned_resources or assigned_resources[ray_resource_key] is None
        else:
            assert ray_resource_key in assigned_resources
            assert assigned_resources[ray_resource_key] == resource

    return c


@udf(return_dtype=daft.DataType.int64())
class AssertResourcesStateful:
    def __init__(self):
        pass

    def __call__(self, c, num_cpus=None, num_gpus=None, memory=None):
        assigned_resources = ray.get_runtime_context().get_assigned_resources()

        for resource, ray_resource_key in [(num_cpus, "CPU"), (num_gpus, "GPU"), (memory, "memory")]:
            if resource is None:
                assert ray_resource_key not in assigned_resources or assigned_resources[ray_resource_key] is None
            else:
                assert ray_resource_key in assigned_resources
                assert assigned_resources[ray_resource_key] == resource

        return c


RAY_VERSION_LT_2 = int(ray.__version__.split(".")[0]) < 2


@pytest.mark.skipif(
    RAY_VERSION_LT_2, reason="The ray.get_runtime_context().get_assigned_resources() was only added in Ray >= 2.0"
)
@pytest.mark.skipif(get_tests_daft_runner_name() not in {"ray"}, reason="requires RayRunner to be in use")
def test_with_column_rayrunner():
    df = daft.from_pydict(DATA).repartition(2)

    assert_resources_parametrized = assert_resources.override_options(num_cpus=1, memory_bytes=1_000_000, num_gpus=None)
    df = df.with_column(
        "resources_ok",
        assert_resources_parametrized(col("id"), num_cpus=1, num_gpus=None, memory=1_000_000),
    )

    df.collect()


@pytest.mark.skipif(
    RAY_VERSION_LT_2, reason="The ray.get_runtime_context().get_assigned_resources() was only added in Ray >= 2.0"
)
@pytest.mark.skipif(get_tests_daft_runner_name() not in {"ray"}, reason="requires RayRunner to be in use")
def test_with_column_folded_rayrunner():
    df = daft.from_pydict(DATA).repartition(2)

    # Because of Projection Folding optimizations, the expected resource request is the max of the three .with_column requests
    expected = dict(num_cpus=1, num_gpus=None, memory=5_000_000)
    df = df.with_column(
        "no_requests",
        assert_resources(col("id"), **expected),
    )

    assert_resources_1 = assert_resources.override_options(num_cpus=1, memory_bytes=5_000_000, num_gpus=None)
    assert_resources_2 = assert_resources.override_options(num_cpus=1, memory_bytes=None, num_gpus=None)
    df = df.with_column(
        "more_memory_request",
        assert_resources_1(col("id"), **expected),
    )
    df = df.with_column(
        "more_cpu_request",
        assert_resources_2(col("id"), **expected),
    )
    df.collect()


@pytest.mark.skipif(
    RAY_VERSION_LT_2, reason="The ray.get_runtime_context().get_assigned_resources() was only added in Ray >= 2.0"
)
@pytest.mark.skipif(get_tests_daft_runner_name() not in {"ray"}, reason="requires RayRunner to be in use")
def test_with_column_rayrunner_class():
    assert_resources = AssertResourcesStateful.with_concurrency(1)

    df = daft.from_pydict(DATA).repartition(2)

    assert_resources_parametrized = assert_resources.override_options(num_cpus=1, memory_bytes=1_000_000, num_gpus=None)
    df = df.with_column(
        "resources_ok",
        assert_resources_parametrized(col("id"), num_cpus=1, num_gpus=None, memory=1_000_000),
    )

    df.collect()


@pytest.mark.skipif(
    RAY_VERSION_LT_2, reason="The ray.get_runtime_context().get_assigned_resources() was only added in Ray >= 2.0"
)
@pytest.mark.skipif(get_tests_daft_runner_name() not in {"ray"}, reason="requires RayRunner to be in use")
def test_with_column_folded_rayrunner_class():
    assert_resources = AssertResourcesStateful.with_concurrency(1)

    df = daft.from_pydict(DATA).repartition(2)

    df = df.with_column(
        "no_requests",
        assert_resources(col("id"), num_cpus=1),  # UDFs have 1 CPU by default
    )

    assert_resources_1 = assert_resources.override_options(num_cpus=1, memory_bytes=5_000_000)
    df = df.with_column(
        "more_memory_request",
        assert_resources_1(col("id"), num_cpus=1, memory=5_000_000),
    )
    df.collect()


###
# GPU tests - can only run if machine has a GPU
###


@udf(return_dtype=daft.DataType.int64(), num_gpus=1)
def assert_num_cuda_visible_devices(c, num_gpus: int = 0):
    cuda_visible_devices_env = os.getenv("CUDA_VISIBLE_DEVICES")
    # Env var not set: program is free to use any number of GPUs
    if cuda_visible_devices_env is None:
        result = len(cuda_visible_devices())
    # Env var set to empty: program has no access to any GPUs
    elif cuda_visible_devices_env == "":
        result = 0
    else:
        result = len(cuda_visible_devices_env.split(","))
    assert result == num_gpus
    return c


@pytest.mark.skipif(get_tests_daft_runner_name() not in {"ray"}, reason="requires RayRunner to be in use")
@pytest.mark.skipif(no_gpu_available(), reason="requires GPUs to be available")
@pytest.mark.parametrize("num_gpus", [None, 1])
def test_with_column_rayrunner_gpu(num_gpus):
    df = daft.from_pydict(DATA).repartition(2)

    assert_num_cuda_visible_devices_parametrized = assert_num_cuda_visible_devices.override_options(num_gpus=num_gpus)
    df = df.with_column(
        "num_cuda_visible_devices",
        assert_num_cuda_visible_devices_parametrized(col("id"), num_gpus=num_gpus if num_gpus is not None else 0),
    )

    df.collect()


@pytest.mark.skipif(get_tests_daft_runner_name() not in {"ray"}, reason="requires RayRunner to be in use")
@pytest.mark.skipif(no_gpu_available(), reason="requires GPUs to be available")
def test_with_column_max_resources_rayrunner_gpu():
    df = daft.from_pydict(DATA).repartition(2)

    # Because of projection folding optimizations, both UDFs should run with num_gpus=1 even though 0_gpu_col requested for 0 GPUs
    assert_num_cuda_visible_devices_0 = assert_num_cuda_visible_devices.override_options(num_gpus=0)
    assert_num_cuda_visible_devices_1 = assert_num_cuda_visible_devices.override_options(num_gpus=1)
    df = df.with_column(
        "0_gpu_col",
        assert_num_cuda_visible_devices_0(col("id"), num_gpus=1),
    )
    df = df.with_column(
        "1_gpu_col",
        assert_num_cuda_visible_devices_1(col("id"), num_gpus=1),
    )

    df.collect()


def test_improper_num_gpus():
    with pytest.raises(ValueError, match="DaftError::ValueError"):

        @udf(return_dtype=daft.DataType.int64(), num_gpus=-1)
        def foo(c):
            return c

    with pytest.raises(ValueError, match="DaftError::ValueError"):

        @udf(return_dtype=daft.DataType.int64(), num_gpus=1.5)
        def foo(c):
            return c

    @udf(return_dtype=daft.DataType.int64())
    def foo(c):
        return c

    with pytest.raises(ValueError, match="DaftError::ValueError"):
        foo = foo.override_options(num_gpus=-1)

    with pytest.raises(ValueError, match="DaftError::ValueError"):
        foo = foo.override_options(num_gpus=1.5)



from __future__ import annotations


def test_with_columns_renamed(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)
    df = df.with_columns_renamed({"sepal_width": "sepal_width_2", "petal_length": "petal_length_2"})
    assert df.column_names == ["sepal_length", "sepal_width_2", "petal_length_2", "petal_width", "variety"]


def test_with_columns_renamed_same_name(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)
    df = df.with_columns_renamed({"sepal_width": "sepal_width"})
    assert df.column_names == ["sepal_length", "sepal_width", "petal_length", "petal_width", "variety"]


def test_with_columns_renamed_empty(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)
    df = df.with_columns_renamed({})
    assert df.column_names == ["sepal_length", "sepal_width", "petal_length", "petal_width", "variety"]


def test_with_columns_renamed_nonexistent_column(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)
    df = df.with_columns_renamed({"sepal_length_2": "sepal_length"})
    assert df.column_names == ["sepal_length", "sepal_width", "petal_length", "petal_width", "variety"]



from __future__ import annotations

import ftfy
import ray
from bs4 import BeautifulSoup

import daft

ray.init()
daft.context.set_runner_ray()


def process_html(html_bytes: bytes | None):
    """Process a single HTML content and extract text."""
    if html_bytes is None:
        return {"extracted_text": None, "text_length": 0}

    try:
        html_str = html_bytes.decode("utf-8")
        if not html_str:
            return {"extracted_text": None, "text_length": 0}

        # Use OWM to extract plain text
        response = BeautifulSoup(html_str, "html.parser").get_text()
        if response is None:
            return {"extracted_text": None, "text_length": 0}

        # Handle potential encoding issues
        try:
            if isinstance(response, str):
                fixed_response = ftfy.fix_text(response)
            else:
                return {"extracted_text": None, "text_length": 0}
        except Exception:
            return {"extracted_text": None, "text_length": 0}

        text_byte_len = len(fixed_response.encode("utf-8"))
        threshold = 100 * (2**20)  # 100MB in bytes

        if text_byte_len > threshold:
            return {"extracted_text": None, "text_length": text_byte_len}
        else:
            return {"extracted_text": fixed_response, "text_length": text_byte_len}

    except Exception:
        return {"extracted_text": None, "text_length": 0}


@daft.udf(
    return_dtype=daft.DataType.struct({"extracted_text": daft.DataType.string(), "text_length": daft.DataType.int32()}),
    batch_size=128,
)
def daft_extract_text(content: daft.Series):
    """Daft UDF to extract text from HTML content using OWM.

    Returns a struct with extracted_text and text_length.
    """
    # Get the config once for all rows
    return [process_html(html_bytes) for html_bytes in content]


# Read 640 CC warc.gz files
daft.read_warc(
    "s3://commoncrawl/crawl-data/CC-MAIN-2018-17/segments/1524125937193.1/warc/CC-MAIN-20180420081400-20180420101400-00*"
).with_column("extracted_text", daft_extract_text(daft.col("warc_content"))).exclude("warc_content").write_parquet(
    "extracted_text"
)



from __future__ import annotations

import pyarrow as pa
import pytest

from daft.expressions import col


@pytest.mark.parametrize(
    "data",
    [
        pa.array([[1, 2], [3, 4], None, []], type=pa.list_(pa.int64())),
        pa.array([[1, 2], [3, 4], None, []], type=pa.large_list(pa.int64())),
    ],
)
def test_explode(make_df, data):
    df = make_df({"nested": data, "sidecar": ["a", "b", "c", "d"]})
    df = df.explode(col("nested"))
    assert df.to_pydict() == {
        "nested": [1, 2, 3, 4, None, None],
        "sidecar": ["a", "a", "b", "b", "c", "d"],
    }


@pytest.mark.parametrize(
    "data",
    [
        pa.array([[1, 2], [3, 4], None, []], type=pa.list_(pa.int64())),
        pa.array([[1, 2], [3, 4], None, []], type=pa.large_list(pa.int64())),
    ],
)
def test_explode_multiple_cols(make_df, data):
    df = make_df({"nested": data, "nested2": data, "sidecar": ["a", "b", "c", "d"]})
    df = df.explode(col("nested"), col("nested2"))
    assert df.to_pydict() == {
        "nested": [1, 2, 3, 4, None, None],
        "nested2": [1, 2, 3, 4, None, None],
        "sidecar": ["a", "a", "b", "b", "c", "d"],
    }


@pytest.mark.parametrize(
    "data_type,sample_data,sidecar,expected_sidecar",
    [
        # List of maps with timestamp values - 4 rows -> 5 rows
        (
            pa.list_(pa.map_(pa.string(), pa.timestamp("us"))),
            [
                [
                    [
                        ("time1", 1717000000000),
                        ("time2", 1717000000001),
                    ]
                ],
                [
                    [
                        ("time1", 1717000000002),
                        ("time2", 1717000000003),
                    ],
                    [
                        ("time3", 1717000000004),
                    ],
                ],
                None,
                [],
            ],
            # initial sidecar
            ["a", "b", "c", "d"],
            # expected sidecar
            ["a", "b", "b", "c", "d"],
        ),
        # List of maps with integer values - 3 rows -> 4 rows
        (
            pa.list_(pa.map_(pa.string(), pa.int64())),
            [
                [
                    [
                        ("score", 95),
                        ("rating", 5),
                    ]
                ],
                [
                    [
                        ("score", 87),
                        ("rating", 4),
                    ],
                    [
                        ("views", 1000),
                    ],
                ],
                [],
            ],
            # initial sidecar
            ["x", "y", "z"],
            # expected sidecar
            ["x", "y", "y", "z"],
        ),
        # List of structs with nested map values - 1 row -> 5 rows
        (
            pa.list_(
                pa.struct(
                    [
                        (
                            "typeMap",
                            pa.map_(pa.string(), pa.map_(pa.string(), pa.list_(pa.string()))),
                        ),
                    ]
                )
            ),
            [
                [
                    {
                        "typeMap": [("genre", [("rock", ["alternative", "indie"])])],
                    },
                    {
                        "typeMap": [("genre", [("pop", ["dance", "electronic"])])],
                    },
                    {
                        "typeMap": [("mood", [("happy", ["upbeat", "energetic"])])],
                    },
                    {
                        "typeMap": [("style", [("classical", ["baroque", "romantic"])])],
                    },
                    {
                        "typeMap": [("era", [("modern", ["contemporary", "experimental"])])],
                    },
                ]
            ],
            # initial sidecar
            ["p"],
            # expected sidecar
            ["p", "p", "p", "p", "p"],
        ),
    ],
)
def test_explode_list_of_maps(make_df, data_type, sample_data, sidecar, expected_sidecar):
    """Test exploding lists of maps with various value types."""
    field = pa.field("data", data_type)
    array = pa.array(sample_data, type=field.type)

    df = make_df({"data": array, "sidecar": sidecar})
    df = df.explode(col("data"))

    result = df.to_pydict()
    # Check that we have the expected number of rows after exploding
    assert len(result["sidecar"]) == len(expected_sidecar)
    assert result["sidecar"] == expected_sidecar


def test_explode_bad_col_type(make_df):
    df = make_df({"a": [1, 2, 3]})
    with pytest.raises(ValueError, match="Input must be a list"):
        df = df.explode(col("a"))



from __future__ import annotations

import pytest

import daft


@pytest.mark.integration()
def test_url_download_local_retry_server(retry_server_s3_config):
    bucket = "80-per-second-rate-limited-gets-bucket"
    data = {"urls": [f"s3://{bucket}/foo{i}" for i in range(100)]}
    df = daft.from_pydict(data)
    df = df.with_column("data", df["urls"].url.download(io_config=retry_server_s3_config, on_error="null"))
    assert df.to_pydict() == {**data, "data": [f"foo{i}".encode() for i in range(100)]}



from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

from daft import Expression, Series, lit
from daft.functions.llm_generate import _vLLMGenerator, llm_generate

vllm = pytest.importorskip("vllm")


@patch("vllm.LLM")
def test_llm_generate_init(mock_llm: MagicMock):
    _vLLMGenerator(model="facebook/opt-125m", generation_config={"temperature": 0.5})
    assert mock_llm.call_count == 1


@patch("vllm.LLM")
def test_llm_generate_init_error(mock_llm: MagicMock):
    mock_llm.side_effect = ImportError("Please install the vllm package to use this provider.")
    with pytest.raises(ImportError):
        _vLLMGenerator(model="facebook/opt-125m", generation_config={"temperature": 0.5})


def test_unsupported_provider_error():
    with pytest.raises(ValueError):
        llm_generate(lit("foo"), provider="unsupported")


def test_returns_expression():
    out = llm_generate(lit("foo"), provider="vllm")
    assert isinstance(out, Expression)
    assert repr(out) == 'py_udf(lit("foo"))'


@patch("vllm.LLM")
@patch("vllm.SamplingParams")
def test_llm_generate_generate(mock_sampling_params: MagicMock, mock_llm: MagicMock):
    # Create mock components
    mock_llm.return_value.generate.return_value = [MagicMock(outputs=[MagicMock(text="This is a mocked response")])]

    llm_generator = _vLLMGenerator(model="facebook/opt-125m", generation_config={"temperature": 0.5})
    series = Series.from_pylist(["This is a test prompt"])
    res = llm_generator(series)
    assert res == ["This is a mocked response"]
    assert mock_sampling_params.call_count == 1
    assert mock_llm.call_count == 1



from __future__ import annotations

import datetime
import decimal

import pyarrow as pa
import pytest

import daft
from daft import DataType, TimeUnit

PYARROW_GE_11_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) >= (11, 0, 0)


@pytest.mark.skipif(
    not PYARROW_GE_11_0_0,
    reason="PyArrow writing to CSV does not have good coverage for all types for versions <11.0.0",
)
@pytest.mark.parametrize(
    ["data", "pa_type", "expected_dtype", "expected_inferred_dtype"],
    [
        ([1, 2, None], pa.int64(), DataType.int64(), DataType.int64()),
        (["a", "b", ""], pa.large_string(), DataType.string(), DataType.string()),
        ([b"a", b"b", b""], pa.large_binary(), DataType.binary(), DataType.string()),
        ([True, False, None], pa.bool_(), DataType.bool(), DataType.bool()),
        ([None, None, None], pa.null(), DataType.null(), DataType.null()),
        (
            [decimal.Decimal("1.23"), decimal.Decimal("1.24"), None],
            pa.decimal128(16, 8),
            DataType.decimal128(16, 8),
            DataType.float64(),
        ),
        ([datetime.date(1994, 1, 1), datetime.date(1995, 1, 1), None], pa.date32(), DataType.date(), DataType.date()),
        (
            [datetime.time(1, 2, 3, 4), datetime.time(5, 6, 7, 8), None],
            pa.time64("us"),
            DataType.time(TimeUnit.us()),
            DataType.time(TimeUnit.us()),
        ),
        (
            [datetime.time(1, 2, 3, 4), datetime.time(5, 6, 7, 8), None],
            pa.time64("ns"),
            DataType.time(TimeUnit.ns()),
            DataType.time(TimeUnit.us()),
        ),
        (
            [datetime.datetime(1994, 1, 1), datetime.datetime(1995, 1, 1), None],
            pa.timestamp("ms"),
            DataType.timestamp(TimeUnit.ms()),
            # NOTE: Seems like the inferred type is seconds because it's written with seconds resolution
            DataType.timestamp(TimeUnit.s()),
        ),
        (
            [datetime.date(1994, 1, 1), datetime.date(1995, 1, 1), None],
            pa.date64(),
            DataType.timestamp(TimeUnit.ms()),
            DataType.timestamp(TimeUnit.s()),
        ),
        (
            [datetime.timedelta(days=1), datetime.timedelta(days=2), None],
            pa.duration("ms"),
            DataType.duration(TimeUnit.ms()),
            # NOTE: Duration ends up being written as int64
            DataType.int64(),
        ),
        # TODO: Verify that these types throw an error when we write dataframes with them
        # ([[1, 2, 3], [], None], pa.large_list(pa.int64()), DataType.list(DataType.int64())),
        # ([[1, 2, 3], [4, 5, 6], None], pa.list_(pa.int64(), list_size=3), DataType.fixed_size_list(DataType.int64(), 3)),
        # ([{"bar": 1}, {"bar": None}, None], pa.struct({"bar": pa.int64()}), DataType.struct({"bar": DataType.int64()})),
    ],
)
def test_roundtrip_simple_arrow_types(tmp_path, data, pa_type, expected_dtype, expected_inferred_dtype):
    before = daft.from_arrow(pa.table({"id": pa.array(range(3)), "foo": pa.array(data, type=pa_type)}))
    before = before.concat(before)
    before.write_csv(str(tmp_path))
    after = daft.read_csv(str(tmp_path))
    assert before.schema()["foo"].dtype == expected_dtype
    assert after.schema()["foo"].dtype == expected_inferred_dtype
    assert before.to_arrow() == after.with_column("foo", after["foo"].cast(expected_dtype)).to_arrow()



from __future__ import annotations

import pytest

from daft import DataType
from daft.expressions import col, lit
from daft.recordbatch import MicroPartition


def test_binary_slice() -> None:
    table = MicroPartition.from_pydict(
        {
            "col": [
                b"foo",
                None,
                b"barbarbar",
                b"quux",
                b"1",
                b"",
                b"Hello\xe2\x98\x83World",  # UTF-8 character in middle
                b"\xf0\x9f\x98\x89test",  # UTF-8 bytes at start
                b"test\xf0\x9f\x8c\x88",  # UTF-8 bytes at end
                b"\xe2\x98\x83\xf0\x9f\x98\x89\xf0\x9f\x8c\x88",  # Multiple UTF-8 sequences
                b"Hello\x00World",  # Null character
                b"\xff\xfe\xfd",  # High bytes
                b"\x00\x01\x02",  # Control characters
            ]
        }
    )
    result = table.eval_expression_list([col("col").binary.slice(0, 5)])
    assert result.to_pydict() == {
        "col": [
            b"foo",
            None,
            b"barba",
            b"quux",
            b"1",
            b"",
            b"Hello",  # Should handle UTF-8 correctly
            b"\xf0\x9f\x98\x89t",  # Should include full UTF-8 sequence
            b"test\xf0",  # Should split UTF-8 sequence
            b"\xe2\x98\x83\xf0\x9f",  # Should split between sequences
            b"Hello",  # Should handle null character
            b"\xff\xfe\xfd",  # Should handle high bytes
            b"\x00\x01\x02",  # Should handle control characters
        ]
    }


@pytest.mark.parametrize(
    "input_data,start_data,length_data,expected_result",
    [
        # Test with column for start position
        (
            [
                b"hello",
                b"world",
                b"test",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd",
            ],
            [1, 0, 2, 5, 1, 4, 1],
            3,
            [b"ell", b"wor", b"st", b"\xe2\x98\x83", b"\x9f\x98\x89", b"\xf0\x9f\x8c", b"\xfe\xfd"],
        ),
        # Test with column for length
        (
            [
                b"hello",
                b"world",
                b"test",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd",
            ],
            1,
            [2, 3, 4, 5, 2, 1, 2],
            [b"el", b"orl", b"est", b"ello\xe2", b"\x9f\x98", b"e", b"\xfe\xfd"],
        ),
        # Test with both start and length as columns
        (
            [
                b"hello",
                b"world",
                b"test",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd",
            ],
            [1, 0, 2, 5, 1, 4, 0],
            [2, 3, 1, 2, 3, 1, 3],
            [b"el", b"wor", b"s", b"\xe2\x98", b"\x9f\x98\x89", b"\xf0", b"\xff\xfe\xfd"],
        ),
        # Test with nulls in start column
        (
            [
                b"hello",
                b"world",
                b"test",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd",
            ],
            [1, None, 2, None, 1, None, 1],
            3,
            [b"ell", None, b"st", None, b"\x9f\x98\x89", None, b"\xfe\xfd"],
        ),
        # Test with nulls in length column
        (
            [
                b"hello",
                b"world",
                b"test",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd",
            ],
            1,
            [2, None, 4, None, 2, None, None],
            [b"el", None, b"est", None, b"\x9f\x98", None, None],
        ),
        # Test with nulls in both columns
        (
            [
                b"hello",
                b"world",
                b"test",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd",
            ],
            [1, None, 2, 5, None, 4, None],
            [2, 3, None, None, 2, None, 2],
            [b"el", None, None, None, None, None, None],
        ),
        # Test with all nulls in start column
        (
            [
                b"hello",
                b"world",
                b"test",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd",
            ],
            [1, None, None, None, None, None, None],
            [2, 3, 1, 2, 3, 1, 3],
            [b"el", None, None, None, None, None, None],
        ),
        # Test with all nulls in length column
        (
            [
                b"hello",
                b"world",
                b"test",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd",
            ],
            [1, 0, 2, 5, 1, 4, 0],
            [2, None, None, None, None, None, None],
            [b"el", None, None, None, None, None, None],
        ),
    ],
)
def test_binary_slice_with_columns(
    input_data: list[bytes | None],
    start_data: list[int | None] | int,
    length_data: list[int | None] | int,
    expected_result: list[bytes | None],
) -> None:
    table_data = {"col": input_data}
    if isinstance(start_data, list):
        table_data["start"] = start_data
        start = col("start")
    else:
        start = start_data

    if isinstance(length_data, list):
        table_data["length"] = length_data
        length = col("length")
    else:
        length = length_data

    table = MicroPartition.from_pydict(table_data)
    result = table.eval_expression_list([col("col").binary.slice(start, length)])
    assert result.to_pydict() == {"col": expected_result}


@pytest.mark.parametrize(
    "input_data,start,length,expected_result",
    [
        # Test start beyond string length
        (
            [
                b"hello",
                b"world",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd\xfc",
            ],
            [10, 20, 15, 10, 10, 5],
            2,
            [b"", b"", b"", b"", b"", b""],
        ),
        # Test start way beyond string length
        (
            [
                b"hello",  # len 5
                b"world",  # len 5
                b"test",  # len 4
                b"\xff\xfe\xfd",  # len 3
            ],
            [100, 1000, 50, 25],
            5,
            [b"", b"", b"", b""],
        ),
        # Test start beyond length with None length
        (
            [
                b"hello",
                b"world",
                b"test",
                b"\xff\xfe\xfd",
            ],
            [10, 20, 15, 8],
            None,
            [b"", b"", b"", b""],
        ),
        # Test zero length
        (
            [
                b"hello",
                b"world",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd\xfc",
            ],
            [1, 0, 5, 0, 4, 2],
            0,
            [b"", b"", b"", b"", b"", b""],
        ),
        # Test very large length
        (
            [
                b"hello",
                b"world",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd\xfc",
            ],
            [0, 1, 5, 0, 4, 1],
            100,
            [b"hello", b"orld", b"\xe2\x98\x83World", b"\xf0\x9f\x98\x89test", b"\xf0\x9f\x8c\x88", b"\xfe\xfd\xfc"],
        ),
        # Test empty strings
        (
            [b"", b"", b"", b""],
            [0, 1, 2, 3],
            3,
            [b"", b"", b"", b""],
        ),
        # Test start + length overflow
        (
            [
                b"hello",
                b"world",
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd\xfc",
            ],
            [2, 3, 5, 0, 4, 2],
            9999999999,
            [b"llo", b"ld", b"\xe2\x98\x83World", b"\xf0\x9f\x98\x89test", b"\xf0\x9f\x8c\x88", b"\xfd\xfc"],
        ),
        # Test UTF-8 and binary sequence boundaries
        (
            [
                b"Hello\xe2\x98\x83World",
                b"\xf0\x9f\x98\x89test",
                b"test\xf0\x9f\x8c\x88",
                b"\xff\xfe\xfd\xfc",
                b"\x00\x01\x02\x03",
            ],
            [4, 0, 3, 1, 2],
            2,
            [b"o\xe2", b"\xf0\x9f", b"t\xf0", b"\xfe\xfd", b"\x02\x03"],
        ),
    ],
)
def test_binary_slice_edge_cases(
    input_data: list[bytes],
    start: list[int],
    length: int,
    expected_result: list[bytes | None],
) -> None:
    table = MicroPartition.from_pydict({"col": input_data, "start": start})
    result = table.eval_expression_list([col("col").binary.slice(col("start"), length)])
    assert result.to_pydict() == {"col": expected_result}


def test_binary_slice_errors() -> None:
    # Test negative start
    table = MicroPartition.from_pydict(
        {"col": [b"hello", b"world", b"Hello\xe2\x98\x83World", b"\xff\xfe\xfd"], "start": [-1, -2, -3, -1]}
    )
    with pytest.raises(Exception, match="DaftError::ComputeError Failed to cast numeric value to target type"):
        table.eval_expression_list([col("col").binary.slice(col("start"), 2)])

    # Test negative length
    table = MicroPartition.from_pydict({"col": [b"hello", b"world", b"Hello\xe2\x98\x83World", b"\xff\xfe\xfd"]})
    with pytest.raises(Exception, match="DaftError::ComputeError Failed to cast numeric value to target type"):
        table.eval_expression_list([col("col").binary.slice(0, -3)])

    # Test both negative
    table = MicroPartition.from_pydict(
        {"col": [b"hello", b"world", b"Hello\xe2\x98\x83World", b"\xff\xfe\xfd"], "start": [-2, -1, -3, -2]}
    )
    with pytest.raises(Exception, match="DaftError::ComputeError Failed to cast numeric value to target type"):
        table.eval_expression_list([col("col").binary.slice(col("start"), -2)])

    # Test negative length in column
    table = MicroPartition.from_pydict(
        {"col": [b"hello", b"world", b"Hello\xe2\x98\x83World", b"\xff\xfe\xfd"], "length": [-2, -3, -4, -2]}
    )
    with pytest.raises(Exception, match="DaftError::ComputeError Failed to cast numeric value to target type"):
        table.eval_expression_list([col("col").binary.slice(0, col("length"))])

    # Test slice with wrong number of arguments (too many)
    table = MicroPartition.from_pydict(
        {"col": [b"hello", b"world"], "start": [1, 2], "length": [2, 3], "extra": [4, 5]}
    )
    with pytest.raises(
        Exception,
        match="(?:ExpressionBinaryNamespace.)?slice\\(\\) takes from 2 to 3 positional arguments but 4 were given",
    ):
        table.eval_expression_list([col("col").binary.slice(col("start"), col("length"), col("extra"))])

    # Test slice with wrong number of arguments (too few)
    table = MicroPartition.from_pydict({"col": [b"hello", b"world"], "start": [1, 2]})
    with pytest.raises(
        Exception, match="(?:ExpressionBinaryNamespace.)?slice\\(\\) missing 1 required positional argument: 'start'"
    ):
        table.eval_expression_list([col("col").binary.slice()])


def test_binary_slice_computed() -> None:
    # Test with computed start index (length - 5)
    table = MicroPartition.from_pydict(
        {
            "col": [
                b"hello world",  # len=11, start=6, expect "world"
                b"python programming",  # len=17, start=12, expect "mming"
                b"data science",  # len=12, start=7, expect "ience"
                b"artificial",  # len=10, start=5, expect "icial"
                b"intelligence",  # len=12, start=7, expect "gence"
                b"Hello\xe2\x98\x83World",  # len=12, start=7, expect "World"
                b"test\xf0\x9f\x98\x89test",  # len=12, start=7, expect "test"
                b"test\xf0\x9f\x8c\x88test",  # len=12, start=7, expect "test"
                b"\xff\xfe\xfd\xfc\xfb",  # len=5, start=0, expect "\xff\xfe\xfd"
            ]
        }
    )
    result = table.eval_expression_list(
        [
            col("col").binary.slice(
                (col("col").binary.length() - 5).cast(DataType.int32()),  # start 5 chars from end
                3,  # take 3 chars
            )
        ]
    )
    assert result.to_pydict() == {
        "col": [b"wor", b"mmi", b"ien", b"ici", b"gen", b"Wor", b"\x89te", b"\x88te", b"\xff\xfe\xfd"]
    }

    # Test with computed length (half of string length)
    table = MicroPartition.from_pydict(
        {
            "col": [
                b"hello world",  # len=11, len/2=5, expect "hello"
                b"python programming",  # len=17, len/2=8, expect "python pr"
                b"data science",  # len=12, len/2=6, expect "data s"
                b"artificial",  # len=10, len/2=5, expect "artif"
                b"intelligence",  # len=12, len/2=6, expect "intell"
                b"Hello\xe2\x98\x83World",  # len=12, len/2=6, expect "Hello\xe2"
                b"test\xf0\x9f\x98\x89test",  # len=12, len/2=6, expect "test\xf0\x9f"
                b"test\xf0\x9f\x8c\x88test",  # len=12, len/2=6, expect "test\xf0\x9f"
                b"\xff\xfe\xfd\xfc\xfb",  # len=5, len/2=2, expect "\xff\xfe"
            ]
        }
    )
    result = table.eval_expression_list(
        [
            col("col").binary.slice(
                0,  # start from beginning
                (col("col").binary.length() / 2).cast(DataType.int32()),  # take half of string
            )
        ]
    )
    assert result.to_pydict() == {
        "col": [
            b"hello",
            b"python pr",
            b"data s",
            b"artif",
            b"intell",
            b"Hello\xe2",
            b"test\xf0\x9f",
            b"test\xf0\x9f",
            b"\xff\xfe",
        ]
    }

    # Test with both computed start and length
    table = MicroPartition.from_pydict(
        {
            "col": [
                b"hello world",  # len=11, start=2, len=3, expect "llo"
                b"python programming",  # len=17, start=3, len=5, expect "hon pr"
                b"data science",  # len=12, start=2, len=4, expect "ta s"
                b"artificial",  # len=10, start=2, len=3, expect "tif"
                b"intelligence",  # len=12, start=2, len=4, expect "tell"
                b"Hello\xe2\x98\x83World",  # len=12, start=2, len=4, expect "llo\xe2"
                b"test\xf0\x9f\x98\x89test",  # len=12, start=2, len=4, expect "st\xf0\x9f"
                b"test\xf0\x9f\x8c\x88test",  # len=12, start=2, len=4, expect "st\xf0\x9f"
                b"\xff\xfe\xfd\xfc\xfb",  # len=5, start=1, len=2, expect "\xfe\xfd"
            ]
        }
    )
    result = table.eval_expression_list(
        [
            col("col").binary.slice(
                (col("col").binary.length() / 5).cast(DataType.int32()),  # start at 1/5 of string
                (col("col").binary.length() / 3).cast(DataType.int32()),  # take 1/3 of string
            )
        ]
    )
    assert result.to_pydict() == {
        "col": [b"llo", b"hon pr", b"ta s", b"tif", b"tell", b"llo\xe2", b"st\xf0\x9f", b"st\xf0\x9f", b"\xfe"]
    }


def test_binary_slice_type_errors() -> None:
    # Test slice with string start type
    table = MicroPartition.from_pydict({"col": [b"hello", b"world"], "start": ["1", "2"]})
    with pytest.raises(
        Exception,
        match="Expects inputs to binary_slice to be binary, integer and integer or null but received Binary, Utf8 and Int32",
    ):
        table.eval_expression_list([col("col").binary.slice(col("start"), 2)])

    # Test slice with float start type
    table = MicroPartition.from_pydict({"col": [b"hello", b"world"], "start": [1.5, 2.5]})
    with pytest.raises(
        Exception,
        match="Expects inputs to binary_slice to be binary, integer and integer or null but received Binary, Float64 and Int32",
    ):
        table.eval_expression_list([col("col").binary.slice(col("start"), 2)])

    # Test slice with boolean start type
    table = MicroPartition.from_pydict({"col": [b"hello", b"world"], "start": [True, False]})
    with pytest.raises(
        Exception,
        match="Expects inputs to binary_slice to be binary, integer and integer or null but received Binary, Boolean and Int32",
    ):
        table.eval_expression_list([col("col").binary.slice(col("start"), 2)])

    # Test slice with binary start type
    table = MicroPartition.from_pydict({"col": [b"hello", b"world"], "start": [b"1", b"2"]})
    with pytest.raises(
        Exception,
        match="Expects inputs to binary_slice to be binary, integer and integer or null but received Binary, Binary and Int32",
    ):
        table.eval_expression_list([col("col").binary.slice(col("start"), 2)])

    # Test slice with null start type
    table = MicroPartition.from_pydict({"col": [b"hello", b"world"], "start": [None, None]})
    with pytest.raises(
        Exception,
        match="Expects inputs to binary_slice to be binary, integer and integer or null but received Binary, Null and Int32",
    ):
        table.eval_expression_list([col("col").binary.slice(col("start"), 2)])


def test_binary_slice_multiple_slices() -> None:
    # Test taking multiple different slices from the same binary data
    table = MicroPartition.from_pydict(
        {
            "col": [
                b"hello",  # Simple ASCII
                b"Hello\xe2\x98\x83World",  # With UTF-8 character
                b"\xf0\x9f\x98\x89test\xf0\x9f\x8c\x88",  # Multiple UTF-8 sequences
                b"\xff\xfe\xfd\xfc\xfb",  # Raw bytes
            ]
        }
    )

    # Get multiple slices
    result = table.eval_expression_list(
        [
            col("col").binary.slice(1, 3).alias("slice1"),  # Middle slice
            col("col").binary.slice(0, 1).alias("slice2"),  # First byte
            col("col").binary.slice(2, 2).alias("slice3"),  # Another middle slice
            col("col")
            .binary.slice((col("col").binary.length().cast(DataType.int64()) - 1), 1)
            .alias("slice4"),  # Last byte
        ]
    )

    assert result.to_pydict() == {
        "slice1": [b"ell", b"ell", b"\x9f\x98\x89", b"\xfe\xfd\xfc"],
        "slice2": [b"h", b"H", b"\xf0", b"\xff"],
        "slice3": [b"ll", b"ll", b"\x98\x89", b"\xfd\xfc"],
        "slice4": [b"o", b"d", b"\x88", b"\xfb"],
    }

    # Test with computed indices
    result = table.eval_expression_list(
        [
            # First half
            col("col").binary.slice(0, (col("col").binary.length() / 2).cast(DataType.int32())).alias("first_half"),
            # Second half
            col("col")
            .binary.slice(
                (col("col").binary.length() / 2).cast(DataType.int32()),
                (col("col").binary.length() / 2).cast(DataType.int32()),
            )
            .alias("second_half"),
            # Middle third
            col("col")
            .binary.slice(
                (col("col").binary.length() / 3).cast(DataType.int32()),
                (col("col").binary.length() / 3).cast(DataType.int32()),
            )
            .alias("middle_third"),
        ]
    )

    assert result.to_pydict() == {
        "first_half": [b"he", b"Hello\xe2", b"\xf0\x9f\x98\x89te", b"\xff\xfe"],
        "second_half": [b"ll", b"\x98\x83Worl", b"st\xf0\x9f\x8c\x88", b"\xfd\xfc"],
        "middle_third": [b"e", b"o\xe2\x98\x83", b"test", b"\xfe"],
    }


@pytest.mark.parametrize(
    "input_data,start,length,expected_result",
    [
        # Test single binary value with multiple start positions
        (
            b"hello",  # Single binary value
            [0, 1, 2, 3, 4, 5],  # Multiple start positions
            2,  # Fixed length
            [b"he", b"el", b"ll", b"lo", b"o", b""],  # Expected slices
        ),
        # Test single binary value with multiple lengths
        (
            b"hello",
            1,  # Fixed start
            [1, 2, 3, 4],  # Multiple lengths
            [b"e", b"el", b"ell", b"ello"],  # Expected slices
        ),
        # Test single binary value with both start and length as arrays
        (
            b"hello",
            [0, 1, 2, 3],  # Multiple starts
            [1, 2, 3, 2],  # Multiple lengths
            [b"h", b"el", b"llo", b"lo"],  # Expected slices
        ),
        # Test with UTF-8 sequences
        (
            # Single UTF-8 string with snowman character
            b"Hello\xe2\x98\x83World",
            [0, 5, 8],  # Multiple starts
            [5, 3, 5],  # Multiple lengths
            # Expected: "Hello", snowman char, "World"
            [b"Hello", b"\xe2\x98\x83", b"World"],
        ),
        # Test with binary data
        (
            b"\xff\xfe\xfd\xfc\xfb",  # Single binary sequence
            [0, 2, 4],  # Multiple starts
            [2, 2, 1],  # Multiple lengths
            [b"\xff\xfe", b"\xfd\xfc", b"\xfb"],  # Expected slices
        ),
        # Test edge cases with single binary value
        (
            b"test",
            [0, 4, 2],  # Start at beginning, end, and middle
            [4, 0, 5],  # Full length, zero length, overflow length
            [b"test", b"", b"st"],  # Expected results
        ),
        # Test with nulls in start/length
        (
            b"hello",
            [0, None, 2, None],  # Mix of valid starts and nulls
            [2, 3, None, None],  # Mix of valid lengths and nulls
            [b"he", None, None, None],  # Expected results with nulls
        ),
    ],
)
def test_binary_slice_broadcasting(
    input_data: bytes,
    start: list[int] | int,
    length: list[int] | int,
    expected_result: list[bytes | None],
) -> None:
    # Create table with slice parameters
    table_data = {}

    # Add start/length columns if they are arrays
    if isinstance(start, list):
        table_data["start"] = start
        start_expr = col("start")
    else:
        start_expr = start

    if isinstance(length, list):
        table_data["length"] = length
        length_expr = col("length")
    else:
        length_expr = length

    # Create table with just the slice parameters
    table = MicroPartition.from_pydict(table_data)

    # Perform slice operation on the raw binary value
    result = table.eval_expression_list([lit(input_data).binary.slice(start_expr, length_expr)])
    assert result.to_pydict() == {"literal": expected_result}



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING, Optional

from daft.daft import PyPartitionField, PyPushdowns, PyRecordBatch, ScanTask
from daft.io.scan import ScanOperator
from daft.logical.schema import Schema
from daft.recordbatch import RecordBatch

if TYPE_CHECKING:
    import lance
    import pyarrow


# TODO support fts and fast_search
def _lancedb_table_factory_function(
    ds: "lance.LanceDataset",
    fragment_ids: Optional[list[int]] = None,
    required_columns: Optional[list[str]] = None,
    filter: Optional["pyarrow.compute.Expression"] = None,
    limit: Optional[int] = None,
) -> Iterator[PyRecordBatch]:
    fragments = [ds.get_fragment(id) for id in (fragment_ids or [])]
    if not fragments:
        raise RuntimeError(f"Unable to find lance fragments {fragment_ids}")
    scanner = ds.scanner(fragments=fragments, columns=required_columns, filter=filter, limit=limit)
    return (RecordBatch.from_arrow_record_batches([rb], rb.schema)._recordbatch for rb in scanner.to_batches())


class LanceDBScanOperator(ScanOperator):
    def __init__(self, ds: "lance.LanceDataset"):
        self._ds = ds

    def name(self) -> str:
        return "LanceDBScanOperator"

    def display_name(self) -> str:
        return f"LanceDBScanOperator({self._ds.uri})"

    def schema(self) -> Schema:
        return Schema.from_pyarrow_schema(self._ds.schema)

    def partitioning_keys(self) -> list[PyPartitionField]:
        return []

    def can_absorb_filter(self) -> bool:
        return False

    def can_absorb_limit(self) -> bool:
        return False

    def can_absorb_select(self) -> bool:
        return False

    def multiline_display(self) -> list[str]:
        return [
            self.display_name(),
            f"Schema = {self.schema()}",
        ]

    def to_scan_tasks(self, pushdowns: PyPushdowns) -> Iterator[ScanTask]:
        required_columns: Optional[list[str]]
        if pushdowns.columns is None:
            required_columns = None
        else:
            filter_required_column_names = pushdowns.filter_required_column_names()
            required_columns = (
                pushdowns.columns
                if filter_required_column_names is None
                else pushdowns.columns + filter_required_column_names
            )

        # TODO: figure out how to translate Pushdowns into LanceDB filters
        filters = None
        fragments = self._ds.get_fragments()
        for i, fragment in enumerate(fragments):
            # TODO: figure out how if we can get this metadata from LanceDB fragments cheaply
            size_bytes = None
            stats = None

            # NOTE: `fragment.count_rows()` should result in 1 IO call for the data file
            # (1 fragment = 1 data file) and 1 more IO call for the deletion file (if present).
            # This could potentially be expensive to perform serially if there are thousands of files.
            # Given that num_rows isn't leveraged for much at the moment, and without statistics
            # we will probably end up materializing the data anyways for any operations, we leave this
            # as None.
            num_rows = None

            yield ScanTask.python_factory_func_scan_task(
                module=_lancedb_table_factory_function.__module__,
                func_name=_lancedb_table_factory_function.__name__,
                func_args=(self._ds, [fragment.fragment_id], required_columns, filters, pushdowns.limit),
                schema=self.schema()._schema,
                num_rows=num_rows,
                size_bytes=size_bytes,
                pushdowns=pushdowns,
                stats=stats,
            )



from __future__ import annotations

import pytest

import daft


@pytest.fixture(scope="function")
def csv_files(tmpdir):
    """Writes 3 CSV files, each of 10 bytes in size, to tmpdir and yield tmpdir."""
    for i in range(3):
        path = tmpdir / f"file.{i}.csv"
        path.write_text("a,b,c\n1,2,", "utf8")  # 10 bytes

    return tmpdir


def test_merge_scan_task_exceed_max(csv_files):
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=0,
        scan_tasks_max_size_bytes=0,
    ):
        df = daft.read_csv(str(csv_files))
        assert (
            df.num_partitions() == 3
        ), "Should have 3 partitions since all merges are more than the maximum (>0 bytes)"


def test_merge_scan_task_below_max(csv_files):
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=11,
        scan_tasks_max_size_bytes=12,
    ):
        df = daft.read_csv(str(csv_files))
        assert (
            df.num_partitions() == 2
        ), "Should have 2 partitions [(CSV1, CSV2), (CSV3)] since the second merge is too large (>22 bytes)"


def test_merge_scan_task_above_min(csv_files):
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=9,
        scan_tasks_max_size_bytes=20,
    ):
        df = daft.read_csv(str(csv_files))
        assert (
            df.num_partitions() == 2
        ), "Should have 2 partitions [(CSV1, CSV2), (CSV3)] since the first merge is above the minimum (>19 bytes)"


def test_merge_scan_task_below_min(csv_files):
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=17,
        scan_tasks_max_size_bytes=20,
    ):
        df = daft.read_csv(str(csv_files))
        assert (
            df.num_partitions() == 1
        ), "Should have 1 partition [(CSV1, CSV2, CSV3)] since both merges are below the minimum and maximum"


def test_merge_scan_task_limit_override(csv_files):
    # A LIMIT operation should override merging of scan tasks, making it only merge up-to the estimated size of the limit
    #
    # With a very small CSV inflation factor, the merger will think that these CSVs provide very few rows of data and will be more aggressive
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=17,
        scan_tasks_max_size_bytes=20,
        csv_inflation_factor=0.1,
    ):
        df = daft.read_csv(str(csv_files)).limit(1)
        assert (
            df.num_partitions() == 1
        ), "Should have 1 partitions [(CSV1, CSV2, CSV3)] since we have a limit 1 but are underestimating the size of data of the CSVs"

    # With a very large CSV inflation factor, the merger will think that these CSVs provide more rows of data and will be more conservative
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=17,
        scan_tasks_max_size_bytes=20,
        csv_inflation_factor=2.0,
    ):
        df = daft.read_csv(str(csv_files)).limit(1)
        assert df.num_partitions() == 3, "Should have 3 partitions [(CSV1, CSV2, CSV3)] since we have a limit 1"


def test_merge_scan_task_up_to_max_sources(csv_files):
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=30,
        scan_tasks_max_size_bytes=30,
        max_sources_per_scan_task=2,
    ):
        df = daft.read_csv(str(csv_files))
        assert (
            df.num_partitions() == 2
        ), "Should have 2 partitions [(CSV1, CSV2), (CSV3)] since the third CSV is too large to merge with the first two, and max_sources_per_scan_task is set to 2"


@pytest.mark.parametrize(
    "min_size_bytes, max_size_bytes",
    [
        (0, 0),
        (0, 10000),
        (10000, 0),
        (10000, 10000),
    ],
)
def test_merge_scan_tasks_does_not_merge_warc(min_size_bytes, max_size_bytes):
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=min_size_bytes,
        scan_tasks_max_size_bytes=max_size_bytes,
    ):
        path = ["tests/assets/example.warc"] * 3
        df = daft.read_warc(path)
        assert df.num_partitions() == 3, "Should have 3 partitions since WARC files are not merged"



from __future__ import annotations

import pytest

from daft import Catalog, Session


@pytest.fixture
def sess():
    cat1 = Catalog.from_pydict(
        name="cat1",
        tables={
            "aa_table": {"x": [1]},
            "bb_table": {"y": [2]},
        },
    )
    cat2 = Catalog.from_pydict(
        name="cat2",
        tables={
            "cc_table": {"x": [1]},
            "dd_table": {"y": [2]},
        },
    )
    sess = Session()
    sess.attach_catalog(cat1)
    sess.attach_catalog(cat2)
    return sess


def test_show_tables(sess):
    res = sess.sql("SHOW TABLES").to_pydict()
    assert "cat1" in res["catalog"]
    assert len(res["table"]) == 2
    assert "aa_table" in res["table"]
    assert "bb_table" in res["table"]


def test_show_tables_all(sess):
    actual = sess.sql("SHOW TABLES").to_pydict()
    expect = {
        "catalog": [
            "cat1",
            "cat1",
        ],
        "namespace": [
            None,
            None,
        ],
        "table": [
            "aa_table",
            "bb_table",
        ],
    }
    assert actual == expect


@pytest.mark.skip(
    "`SHOW TABLES LIKE` is not properly supported yet. See https://github.com/Eventual-Inc/Daft/issues/4461"
)
def test_show_tables_with_pattern(sess):
    # test: show tables in the current catalog and current namespace matching the pattern
    res = sess.sql("SHOW TABLES LIKE 'aa'").to_pydict()
    assert len(res["table"]) == 1
    assert "aa_table" in res["table"]


def test_show_tables_in_catalog(sess):
    # test: show tables in catalog `cat2`
    res = sess.sql("SHOW TABLES FROM cat2").to_pydict()
    assert "cat2" in res["catalog"]
    assert len(res["table"]) == 2
    assert "cc_table" in res["table"]
    assert "dd_table" in res["table"]


@pytest.mark.skip(
    "`SHOW TABLES LIKE` is not properly supported yet. See https://github.com/Eventual-Inc/Daft/issues/4461"
)
def test_show_tables_in_catalog_with_pattern(sess):
    # test: show tables in catalog `cat2` matching the pattern
    res = sess.sql("SHOW TABLES FROM cat2 LIKE 'cc'").to_pydict()
    assert "cat2" in res["catalog"]
    assert len(res["table"]) == 1
    assert "cc_table" in res["table"]


@pytest.mark.skip("FROM <catalog>.<namespace> not supported in sqlparser-rs")
def test_show_tables_in_catalog_namespace(sess):
    # test: show tables in catalog `cat2` and namespace `my_namespace`
    result = sess.sql("SHOW TABLES FROM cat2.my_namespace")
    tables = result.to_pydict()["table"]
    assert len(tables) == 2  # Only tables in my_namespace
    assert "foo_table" in tables
    assert "bar_table" in tables


@pytest.mark.skip("FROM <catalog>.<namespace> not supported in sqlparser-rs")
def test_show_tables_in_catalog_namespace_with_pattern(sess):
    # test: show tables in catalog `cat2` and namespace `my_namespace` matching the pattern
    result = sess.sql("SHOW TABLES FROM cat2.my_namespace LIKE 'foo'")
    tables = result.to_pydict()["table"]
    assert len(tables) == 1
    assert "foo_table" in tables
    assert "bar_table" not in tables



from __future__ import annotations

from pyspark.sql import Row


def test_distinct(spark_session):
    # Create simple DataFrame with single column
    data = [(1,), (2,), (1,)]
    df = spark_session.createDataFrame(data, ["id"]).distinct()

    assert df.count() == 2, "DataFrame should have 2 rows"

    assert df.sort().collect() == [Row(id=1), Row(id=2)], "DataFrame should contain expected values"



from __future__ import annotations

import numpy as np
import pyarrow as pa
import pytest

import daft
from daft.recordbatch import MicroPartition

MODES = ["L", "LA", "RGB", "RGBA"]
MODE_TO_NP_DTYPE = {
    "L": np.uint8,
    "LA": np.uint8,
    "RGB": np.uint8,
    "RGBA": np.uint8,
    "L16": np.uint16,
    "LA16": np.uint16,
    "RGB16": np.uint16,
    "RGBA16": np.uint16,
    "RGB32F": np.float32,
    "RGBA32F": np.float32,
}
MODE_TO_NUM_CHANNELS = {
    "L": 1,
    "LA": 2,
    "RGB": 3,
    "RGBA": 4,
    "L16": 1,
    "LA16": 2,
    "RGB16": 3,
    "RGBA16": 4,
    "RGB32F": 3,
    "RGBA32F": 4,
}


def test_image_crop_mixed_shape_same_mode(mixed_shape_data_fixture):
    table = MicroPartition.from_pydict({"images": mixed_shape_data_fixture})
    result = table.eval_expression_list([daft.col("images").image.crop((1, 1, 1, 1))])
    result = result.to_pydict()

    def crop(arr):
        return arr[1:2, 1:2, :]

    expected = [crop(arr) if arr is not None else None for arr in mixed_shape_data_fixture.to_pylist()]
    np.testing.assert_equal(result["images"], expected)


def test_image_crop_mixed_shape_same_mode_crop_col(mixed_shape_data_fixture):
    # TODO(jay): Need to fix nested casts -- for now we workaround by creating a nested uint32
    bboxes = pa.array([[1, 1, 1, 1], None, None], type=pa.list_(pa.uint32()))
    table = MicroPartition.from_pydict(
        {
            "images": mixed_shape_data_fixture,
            "bboxes": bboxes,
        }
    )
    result = table.eval_expression_list([daft.col("images").image.crop(daft.col("bboxes"))])
    result = result.to_pydict()

    def crop(arr):
        return arr[1:2, 1:2, :]

    expected = [crop(mixed_shape_data_fixture.to_pylist()[0]), None, None]
    np.testing.assert_equal(result["images"], expected)


def test_image_crop_fixed_shape_same_mode(fixed_shape_data_fixture):
    table = MicroPartition.from_pydict({"images": fixed_shape_data_fixture})
    result = table.eval_expression_list([daft.col("images").image.crop((1, 1, 1, 1))])
    result = result.to_pydict()

    def crop(arr):
        return arr[1:2, 1:2, :]

    expected = [crop(arr) if arr is not None else None for arr in fixed_shape_data_fixture.to_pylist()]
    np.testing.assert_equal(result["images"], expected)


def test_image_crop_fixed_shape_same_mode_crop_col(fixed_shape_data_fixture):
    # TODO(jay): Need to fix nested casts -- for now we workaround by creating a nested uint32
    bboxes = pa.array([[1, 1, 1, 1], None, None], type=pa.list_(pa.uint32()))
    table = MicroPartition.from_pydict(
        {
            "images": fixed_shape_data_fixture,
            "bboxes": bboxes,
        }
    )
    result = table.eval_expression_list([daft.col("images").image.crop(daft.col("bboxes"))])
    result = result.to_pydict()

    def crop(arr):
        return arr[1:2, 1:2, :]

    expected = [crop(fixed_shape_data_fixture.to_pylist()[0]), None, None]
    np.testing.assert_equal(result["images"], expected)


def test_bad_expr_input():
    table = MicroPartition.from_pydict({"x": [1, 2, 3], "y": ["a", "b", "c"]})

    # Test bad Expression calls
    with pytest.raises(ValueError):
        daft.col("x").image.crop("foo")

    with pytest.raises(ValueError):
        daft.col("x").image.crop([1, 2, 3, 4, 5])

    with pytest.raises(ValueError):
        daft.col("x").image.crop([1.0, 2.0, 3.0, 4.0])

    # Test calling on bad types
    with pytest.raises(ValueError):
        table.eval_expression_list([daft.col("x").image.crop((1, 2, 3, 4))])



from __future__ import annotations

import random
import tempfile
from contextlib import contextmanager
from functools import partial
from typing import Callable

import numpy as np
import pyarrow as pa
import pytest

import daft
from daft.io._generator import read_generator
from daft.recordbatch.recordbatch import RecordBatch
from tests.conftest import get_tests_daft_runner_name


def generate(num_rows: int, bytes_per_row: int):
    data = {
        "ints": np.random.randint(0, num_rows, num_rows, dtype=np.uint64),
        "bytes": pa.array(
            [np.random.bytes(bytes_per_row) for _ in range(num_rows)],
            type=pa.binary(bytes_per_row),
        ),
    }
    yield RecordBatch.from_pydict(data)


def generator(
    num_partitions: int,
    num_rows_fn: Callable[[], int],
    bytes_per_row_fn: Callable[[], int],
):
    for _ in range(num_partitions):
        num_rows = num_rows_fn()
        bytes_per_row = bytes_per_row_fn()
        yield partial(generate, num_rows, bytes_per_row)


@pytest.fixture(scope="function")
def pre_shuffle_merge_ctx():
    """Fixture that provides a context manager for pre-shuffle merge testing."""

    def _ctx(threshold: int | None = None):
        return daft.execution_config_ctx(shuffle_algorithm="pre_shuffle_merge", pre_shuffle_merge_threshold=threshold)

    return _ctx


@pytest.fixture(scope="function")
def flight_shuffle_ctx():
    """Fixture that provides a context manager for flight shuffle testing with a temporary directory."""

    @contextmanager
    def _ctx():
        # Create a temporary directory that automatically cleans up
        with tempfile.TemporaryDirectory() as temp_dir:
            # Use the temporary directory for flight shuffle
            with daft.execution_config_ctx(shuffle_algorithm="flight_shuffle", flight_shuffle_dirs=[temp_dir]) as ctx:
                yield ctx

    return _ctx


@pytest.mark.skipif(
    get_tests_daft_runner_name() != "ray",
    reason="shuffle tests are meant for the ray runner",
)
@pytest.mark.parametrize(
    "input_partitions, output_partitions",
    [(100, 100), (100, 1), (100, 50), (100, 200)],
)
def test_pre_shuffle_merge_small_partitions(pre_shuffle_merge_ctx, input_partitions, output_partitions):
    """Test that pre-shuffle merge is working for small partitions less than the memory threshold."""

    def num_rows_fn():
        return output_partitions

    def bytes_per_row_fn():
        return 1

    threshold = None

    with pre_shuffle_merge_ctx(threshold):
        df = (
            read_generator(
                generator(input_partitions, num_rows_fn, bytes_per_row_fn),
                schema=daft.Schema._from_field_name_and_types(
                    [
                        ("ints", daft.DataType.uint64()),
                        ("bytes", daft.DataType.binary()),
                    ]
                ),
            )
            .repartition(output_partitions, "ints")
            .collect()
        )
        assert len(df) == input_partitions * output_partitions


@pytest.mark.skipif(
    get_tests_daft_runner_name() != "ray",
    reason="shuffle tests are meant for the ray runner",
)
@pytest.mark.parametrize(
    "input_partitions, output_partitions",
    [(100, 100), (100, 1), (100, 50), (100, 200)],
)
def test_pre_shuffle_merge_big_partitions(pre_shuffle_merge_ctx, input_partitions, output_partitions):
    """Test that pre-shuffle merge is working for big partitions greater than the threshold."""

    def num_rows_fn():
        return output_partitions

    def bytes_per_row_fn():
        return 200

    threshold = 1

    with pre_shuffle_merge_ctx(threshold):
        df = (
            read_generator(
                generator(input_partitions, num_rows_fn, bytes_per_row_fn),
                schema=daft.Schema._from_field_name_and_types(
                    [
                        ("ints", daft.DataType.uint64()),
                        ("bytes", daft.DataType.binary()),
                    ]
                ),
            )
            .repartition(output_partitions, "ints")
            .collect()
        )
        assert len(df) == input_partitions * output_partitions


@pytest.mark.skipif(
    get_tests_daft_runner_name() != "ray",
    reason="shuffle tests are meant for the ray runner",
)
@pytest.mark.parametrize(
    "input_partitions, output_partitions",
    [(100, 100), (100, 1), (100, 50), (100, 200)],
)
def test_pre_shuffle_merge_randomly_sized_partitions(pre_shuffle_merge_ctx, input_partitions, output_partitions):
    """Test that pre-shuffle merge is working for randomly sized partitions."""

    def num_rows_fn():
        return output_partitions

    def bytes_per_row_fn():
        return random.randint(1, output_partitions // 2 + 1)

    # We want some partitions that are small, and some that are big. We want to cap the big ones to be around half of the threshold.
    threshold = output_partitions * (8 + output_partitions)

    with pre_shuffle_merge_ctx(threshold):
        df = (
            read_generator(
                generator(input_partitions, num_rows_fn, bytes_per_row_fn),
                schema=daft.Schema._from_field_name_and_types(
                    [
                        ("ints", daft.DataType.uint64()),
                        ("bytes", daft.DataType.binary()),
                    ]
                ),
            )
            .repartition(output_partitions, "ints")
            .collect()
        )
        assert len(df) == input_partitions * output_partitions


@pytest.mark.skipif(
    get_tests_daft_runner_name() != "ray",
    reason="shuffle tests are meant for the ray runner",
)
@pytest.mark.parametrize(
    "input_partitions, output_partitions",
    [(100, 100), (100, 1), (100, 50), (100, 200)],
)
def test_flight_shuffle(flight_shuffle_ctx, input_partitions, output_partitions):
    """Test that flight shuffle is working."""

    def num_rows_fn():
        return output_partitions

    def bytes_per_row_fn():
        return 200

    with flight_shuffle_ctx():
        df = (
            read_generator(
                generator(input_partitions, num_rows_fn, bytes_per_row_fn),
                schema=daft.Schema._from_field_name_and_types(
                    [
                        ("ints", daft.DataType.uint64()),
                        ("bytes", daft.DataType.fixed_size_binary(200)),
                    ]
                ),
            )
            .repartition(output_partitions, "ints")
            .collect()
        )
        assert len(df) == input_partitions * output_partitions



from __future__ import annotations

import pytest


def test_create_or_replace_temp_view(spark_session):
    df = spark_session.createDataFrame([(1, "foo")], ["id", "name"])
    try:
        df.createOrReplaceTempView("test_view")
    except Exception as e:
        pytest.fail(f"createOrReplaceTempView failed: {e}")


def test_sql(spark_session):
    df = spark_session.createDataFrame([(1, "foo")], ["id", "name"])
    df.createOrReplaceTempView("test_view")
    try:
        result = spark_session.sql("SELECT * FROM test_view")
    except Exception as e:
        pytest.fail(f"sql failed: {e}")
    assert result.collect() == [(1, "foo")]



from __future__ import annotations

import sys

import pytest
from fsspec.implementations.local import LocalFileSystem

import daft
from benchmarking.tpch import answers, data_generation

if sys.platform == "win32":
    pytest.skip(allow_module_level=True)


@pytest.fixture(scope="module")
def get_df(gen_tpch):
    csv_files_location, _ = gen_tpch

    def _get_df(tbl_name: str):
        # TODO (jay): Perhaps we should use Parquet here instead similar to benchmarking and get rid of this CSV parsing stuff?
        local_fs = LocalFileSystem()
        # Used chunked files if found
        nonchunked_filepath = f"{csv_files_location}/{tbl_name}.tbl"
        chunked_filepath = nonchunked_filepath + ".*"
        try:
            local_fs.expand_path(chunked_filepath)
            fp = chunked_filepath
        except FileNotFoundError:
            fp = nonchunked_filepath

        df = daft.read_csv(
            fp,
            has_headers=False,
            delimiter="|",
        )
        df = df.select(
            *[
                daft.col(autoname).alias(colname)
                for autoname, colname in zip(df.column_names, data_generation.SCHEMA[tbl_name])
            ]
        )
        return df

    return _get_df


def test_tpch_q1(tmp_path, check_answer, get_df):
    daft_df = answers.q1(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 1, tmp_path)


def test_tpch_q2(tmp_path, check_answer, get_df):
    daft_df = answers.q2(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 2, tmp_path)


def test_tpch_q3(tmp_path, check_answer, get_df):
    daft_df = answers.q3(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 3, tmp_path)


def test_tpch_q4(tmp_path, check_answer, get_df):
    daft_df = answers.q4(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 4, tmp_path)


def test_tpch_q5(tmp_path, check_answer, get_df):
    daft_df = answers.q5(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 5, tmp_path)


def test_tpch_q6(tmp_path, check_answer, get_df):
    daft_df = answers.q6(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 6, tmp_path)


def test_tpch_q7(tmp_path, check_answer, get_df):
    daft_df = answers.q7(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 7, tmp_path)


def test_tpch_q8(tmp_path, check_answer, get_df):
    daft_df = answers.q8(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 8, tmp_path)


def test_tpch_q9(tmp_path, check_answer, get_df):
    daft_df = answers.q9(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 9, tmp_path)


def test_tpch_q10(tmp_path, check_answer, get_df):
    daft_df = answers.q10(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 10, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q11(tmp_path, check_answer, get_df):
    daft_df = answers.q11(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 11, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q12(tmp_path, check_answer, get_df):
    daft_df = answers.q12(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 12, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q13(tmp_path, check_answer, get_df):
    daft_df = answers.q13(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 13, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q14(tmp_path, check_answer, get_df):
    daft_df = answers.q14(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 14, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q15(tmp_path, check_answer, get_df):
    daft_df = answers.q15(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 15, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q16(tmp_path, check_answer, get_df):
    daft_df = answers.q16(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 16, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q17(tmp_path, check_answer, get_df):
    daft_df = answers.q17(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 17, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q18(tmp_path, check_answer, get_df):
    daft_df = answers.q18(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 18, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q19(tmp_path, check_answer, get_df):
    daft_df = answers.q19(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 19, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q20(tmp_path, check_answer, get_df):
    daft_df = answers.q20(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 20, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q21(tmp_path, check_answer, get_df):
    daft_df = answers.q21(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 21, tmp_path)


@pytest.mark.skip(reason="Running all TPC-H queries is too slow")
def test_tpch_q22(tmp_path, check_answer, get_df):
    daft_df = answers.q22(get_df)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 22, tmp_path)



from __future__ import annotations

from itertools import product

import pytest

import daft
from daft.daft import JsonConvertOptions
from daft.recordbatch import MicroPartition

PRED_PUSHDOWN_FILES = [
    "s3://daft-public-data/test_fixtures/json-dev/sampled-tpch.jsonl",
    "tests/assets/sampled-tpch.jsonl",
]


@pytest.mark.integration()
@pytest.mark.parametrize(
    "path, pred, limit",
    product(
        PRED_PUSHDOWN_FILES,
        [daft.col("L_ORDERKEY") == 7, daft.col("L_ORDERKEY") == 10000, daft.lit(True), None],
        [None, 1, 1000],
    ),
)
def test_json_filter_pushdowns(path, pred, limit, aws_public_s3_config):
    with_pushdown = MicroPartition.read_json(
        path,
        io_config=aws_public_s3_config,
        convert_options=JsonConvertOptions(limit=limit, predicate=pred._expr if pred is not None else None),
        parse_options=None,
        read_options=None,
    )
    after = MicroPartition.read_json(
        path, io_config=aws_public_s3_config, convert_options=None, parse_options=None, read_options=None
    )
    if pred is not None:
        after = after.filter([pred])
    if limit is not None:
        after = after.head(limit)
    assert with_pushdown.to_arrow() == after.to_arrow()



from __future__ import annotations

import random

import pandas as pd
import pytest

from daft import Window, col
from tests.conftest import assert_df_equals, get_tests_daft_runner_name

# from daft.expressions import count, max, mean, min, sum


pytestmark = pytest.mark.skipif(
    get_tests_daft_runner_name() != "native", reason="Window tests only run on native runner"
)


def test_single_partition_sum(make_df):
    """Stage: PARTITION BY-Only Window Aggregations.

    Test sum over a single partition column.
    """
    df = make_df({"category": ["B", "A", "C", "A", "B", "C", "A", "B"], "value": [10, 5, 15, 8, 12, 6, 9, 7]})

    window = Window().partition_by("category")
    result = df.select(
        col("category"),
        col("value"),
        # sum("value").over(window).alias("sum"), # TODO: Support .over() directly on expressions
        col("value").sum().over(window).alias("sum"),
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "value": [5, 8, 9, 10, 12, 7, 15, 6],
        "sum": [22, 22, 22, 29, 29, 29, 21, 21],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()))


def test_single_partition_min(make_df):
    """Test min over a single partition column."""
    df = make_df({"category": ["B", "A", "C", "A", "B", "C", "A", "B"], "value": [10, 5, 15, 8, 12, 6, 9, 7]})

    window = Window().partition_by("category")
    result = df.select(
        col("category"),
        col("value"),
        col("value").min().over(window).alias("min"),
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "value": [5, 8, 9, 10, 12, 7, 15, 6],
        "min": [5, 5, 5, 7, 7, 7, 6, 6],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()))


def test_single_partition_max(make_df):
    """Test max over a single partition column."""
    df = make_df({"category": ["B", "A", "C", "A", "B", "C", "A", "B"], "value": [10, 5, 15, 8, 12, 6, 9, 7]})

    window = Window().partition_by("category")
    result = df.select(
        col("category"),
        col("value"),
        col("value").max().over(window).alias("max"),
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "value": [5, 8, 9, 10, 12, 7, 15, 6],
        "max": [9, 9, 9, 12, 12, 12, 15, 15],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()))


def test_single_partition_mean(make_df):
    """Test mean over a single partition column."""
    df = make_df({"category": ["B", "A", "C", "A", "B", "C", "A", "B"], "value": [10, 5, 15, 8, 12, 6, 9, 7]})

    window = Window().partition_by("category")
    result = df.select(
        col("category"),
        col("value"),
        col("value").mean().over(window).alias("mean"),
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "value": [5, 8, 9, 10, 12, 7, 15, 6],
        "mean": [
            7.333333333333333,
            7.333333333333333,
            7.333333333333333,
            9.666666666666666,
            9.666666666666666,
            9.666666666666666,
            10.5,
            10.5,
        ],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()))


def test_single_partition_count(make_df):
    """Test count over a single partition column."""
    df = make_df({"category": ["B", "A", "C", "A", "B", "C", "A", "B"], "value": [10, 5, 15, 8, 12, 6, 9, 7]})

    window = Window().partition_by("category")
    result = df.select(
        col("category"),
        col("value"),
        col("value").count().over(window).alias("count"),
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "value": [5, 8, 9, 10, 12, 7, 15, 6],
        "count": [3, 3, 3, 3, 3, 3, 2, 2],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()), check_dtype=False)


def test_multiple_partition_columns(make_df):
    """Test sum over multiple partition columns."""
    df = make_df(
        {
            "category": ["B", "A", "C", "A", "B", "C", "A", "B"],
            "group": [1, 1, 2, 1, 2, 2, 2, 1],
            "value": [10, 5, 15, 8, 12, 6, 9, 7],
        }
    )

    window = Window().partition_by(["category", "group"])
    result = df.select(
        col("category"),
        col("group"),
        col("value"),
        col("value").sum().over(window).alias("sum"),
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "group": [1, 1, 2, 1, 2, 1, 2, 2],
        "value": [5, 8, 9, 10, 12, 7, 15, 6],
        "sum": [13, 13, 9, 17, 12, 17, 21, 21],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()))


def test_similar_partition_keys(make_df):
    """Test partition keys with similar string/number combinations that might cause hash collisions."""
    df = make_df(
        {
            "key1": ["A1", "B2", "C3", "D4", "E5", "A2", "B1", "C4", "D3", "E2"],
            "key2": ["X1", "Y2", "Z3", "X2", "Y3", "Z1", "X3", "Y1", "Z2", "X4"],
            "value": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        }
    )

    window = Window().partition_by(["key1", "key2"])
    result = df.select(
        col("key1"),
        col("key2"),
        col("value"),
        col("value").sum().over(window).alias("sum"),
    ).collect()

    expected_dict = result.to_pydict()

    for i in range(len(expected_dict["key1"])):
        assert (
            expected_dict["sum"][i] == expected_dict["value"][i]
        ), f"Hash collision detected for keys {expected_dict['key1'][i]}/{expected_dict['key2'][i]}"


def test_null_partition_values(make_df):
    """Test window functions with null values in partition columns."""
    df = make_df(
        {
            "category": ["A", "A", None, "B", "B", None, "C", None],
            "value": [5, 8, 9, 10, 12, 7, 15, 6],
        }
    )

    window = Window().partition_by("category")
    result = df.select(
        col("category"),
        col("value"),
        col("value").sum().over(window).alias("sum"),
    ).collect()

    result_dict = result.to_pydict()

    for category in set(result_dict["category"]):
        if category is None:
            null_indices = [i for i, cat in enumerate(result_dict["category"]) if cat is None]
            null_values = [result_dict["value"][i] for i in null_indices]
            expected_sum = sum(null_values)
            actual_sums = [result_dict["sum"][i] for i in null_indices]
            assert all(
                sum == expected_sum for sum in actual_sums
            ), f"Incorrect sum for null category: {actual_sums} != {expected_sum}"
        else:
            cat_indices = [i for i, cat in enumerate(result_dict["category"]) if cat == category]
            cat_values = [result_dict["value"][i] for i in cat_indices]
            expected_sum = sum(cat_values)
            actual_sums = [result_dict["sum"][i] for i in cat_indices]
            assert all(
                sum == expected_sum for sum in actual_sums
            ), f"Incorrect sum for category {category}: {actual_sums} != {expected_sum}"


def test_multiple_window_functions(make_df):
    """Test multiple window functions in the same query."""
    df = make_df({"category": ["B", "A", "C", "A", "B", "C", "A", "B"], "value": [10, 5, 15, 8, 12, 6, 9, 7]})

    window = Window().partition_by("category")
    result = df.select(
        col("category"),
        col("value"),
        col("value").sum().over(window).alias("sum"),
        col("value").min().over(window).alias("min"),
        col("value").max().over(window).alias("max"),
        col("value").mean().over(window).alias("mean"),
        col("value").count().over(window).alias("count"),
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "value": [5, 8, 9, 10, 12, 7, 15, 6],
        "sum": [22, 22, 22, 29, 29, 29, 21, 21],
        "min": [5, 5, 5, 7, 7, 7, 6, 6],
        "max": [9, 9, 9, 12, 12, 12, 15, 15],
        "mean": [
            7.333333333333333,
            7.333333333333333,
            7.333333333333333,
            9.666666666666666,
            9.666666666666666,
            9.666666666666666,
            10.5,
            10.5,
        ],
        "count": [3, 3, 3, 3, 3, 3, 2, 2],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()), check_dtype=False)


def test_many_partitions(make_df):
    """Test window functions with a large number of partitions."""
    num_partitions = 100
    categories = [f"category_{i}" for i in range(num_partitions)]
    values = list(range(num_partitions))

    df = make_df({"category": categories, "value": values})

    window = Window().partition_by("category")
    result = df.select(
        col("category"),
        col("value"),
        col("value").sum().over(window).alias("sum"),
    ).collect()

    result_dict = result.to_pydict()

    for i in range(len(result_dict["category"])):
        assert (
            result_dict["sum"][i] == result_dict["value"][i]
        ), f"Expected sum equal to value for single-row partition {result_dict['category'][i]}"


def test_window_mean_minus_value(make_df):
    """Test arithmetic with window mean and value."""
    df = make_df({"category": ["B", "A", "C", "A", "B", "C", "A", "B"], "value": [10, 5, 15, 8, 12, 6, 9, 7]})

    window = Window().partition_by("category")
    result = df.select(
        col("category"),
        col("value"),
        (col("value").mean().over(window) - col("value")).alias("mean_minus_value"),
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "value": [5, 8, 9, 7, 10, 12, 6, 15],
        "mean_minus_value": [
            2.333333333333333,
            -0.666666666666667,
            -1.666666666666667,
            2.666666666666666,
            -0.3333333333333339,
            -2.333333333333334,
            4.5,
            -4.5,
        ],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()))


def test_window_value_over_sum(make_df):
    """Test division with value and window sum."""
    df = make_df({"category": ["B", "A", "C", "A", "B", "C", "A", "B"], "value": [10, 5, 15, 8, 12, 6, 9, 7]})

    window = Window().partition_by("category")
    result = df.select(
        col("category"),
        col("value"),
        (col("value") / col("value").sum().over(window)).alias("value_over_sum"),
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "value": [5, 8, 9, 7, 10, 12, 6, 15],
        "value_over_sum": [
            0.22727272727272727,
            0.36363636363636365,
            0.4090909090909091,
            0.2413793103448276,
            0.3448275862068966,
            0.41379310344827586,
            0.2857142857142857,
            0.7142857142857143,
        ],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()))


def test_factor_window_expression(make_df):
    """Test factor window expression."""
    df = make_df({"category": ["B", "A", "C", "A", "B", "C", "A", "B"], "value": [10, 5, 15, 8, 12, 6, 9, 7]})

    window = Window().partition_by("category")
    result = df.select(
        col("category"),
        col("value"),
        col("value").sum().over(window).alias("sum"),
        (col("value") / col("value").sum().over(window)).alias("value_over_sum"),
        col("value").max().over(window).alias("max"),
        (col("value") - col("value").max().over(window)).alias("value_minus_max"),
        col("value").min().over(window).alias("min"),
        (col("value") - col("value").min().over(window)).alias("value_minus_min"),
        (col("value").max().over(window) - col("value").min().over(window)).alias("max_minus_min"),
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "value": [5, 8, 9, 7, 10, 12, 6, 15],
        "sum": [22, 22, 22, 29, 29, 29, 21, 21],
        "value_over_sum": [
            0.22727272727272727,
            0.36363636363636365,
            0.4090909090909091,
            0.2413793103448276,
            0.3448275862068966,
            0.41379310344827586,
            0.2857142857142857,
            0.7142857142857143,
        ],
        "max": [9, 9, 9, 12, 12, 12, 15, 15],
        "value_minus_max": [-4, -1, 0, -5, -2, 0, -9, 0],
        "min": [5, 5, 5, 7, 7, 7, 6, 6],
        "value_minus_min": [0, 3, 4, 0, 3, 5, 0, 9],
        "max_minus_min": [4, 4, 4, 5, 5, 5, 9, 9],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()))


def test_partition_by_with_expressions(make_df):
    """Test window functions with expressions in partition_by clause."""
    df = make_df({"num": [1, 2, 3, 4, 5, 6, 7, 8], "value": [10, 20, 30, 40, 50, 60, 70, 80]})

    df = df.with_column("parity", col("num") % 2)

    parity_expr = col("parity")
    window = Window().partition_by(parity_expr)

    result = df.select(
        col("num"),
        col("value"),
        col("parity"),
        col("value").sum().over(window).alias("sum"),
    ).collect()

    result_dict = result.to_pydict()

    odd_values = [result_dict["value"][i] for i, p in enumerate(result_dict["parity"]) if p == 1]
    even_values = [result_dict["value"][i] for i, p in enumerate(result_dict["parity"]) if p == 0]

    odd_sum = sum(odd_values)
    even_sum = sum(even_values)

    for i, parity in enumerate(result_dict["parity"]):
        expected_sum = odd_sum if parity == 1 else even_sum
        assert result_dict["sum"][i] == expected_sum, f"Row {i} has incorrect sum for parity {parity}"


def test_multiple_window_partitions(make_df):
    """Test multiple window functions with different partition keys using random numbers.

    Creates a dataset with 18 rows (2 rows each for A1, A2, A3, B1, B2, B3, C1, C2, C3)
    and verifies sums across different partition keys (A/B/C and 1/2/3).
    """
    random.seed(42)

    data = []
    for letter in ["A", "B", "C"]:
        for num in ["1", "2", "3"]:
            for _ in range(3):
                data.append({"letter": letter, "num": num, "value": random.randint(1, 100)})

    df = make_df(data)

    letter_window = Window().partition_by("letter")
    num_window = Window().partition_by("num")
    combined_window = Window().partition_by(["letter", "num"])

    result = df.select(
        col("letter"),
        col("num"),
        col("value"),
        col("value").sum().over(letter_window).alias("letter_sum"),
        col("value").sum().over(num_window).alias("num_sum"),
        col("value").sum().over(combined_window).alias("combined_sum"),
    ).collect()

    result_dict = result.to_pydict()

    for letter in ["A", "B", "C"]:
        letter_indices = [i for i, ltr in enumerate(result_dict["letter"]) if ltr == letter]
        letter_values = [result_dict["value"][i] for i in letter_indices]
        expected_letter_sum = sum(letter_values)
        actual_letter_sums = [result_dict["letter_sum"][i] for i in letter_indices]
        assert all(
            sum == expected_letter_sum for sum in actual_letter_sums
        ), f"Incorrect sum for letter {letter}: {actual_letter_sums} != {expected_letter_sum}"

    for num in ["1", "2", "3"]:
        num_indices = [i for i, n in enumerate(result_dict["num"]) if n == num]
        num_values = [result_dict["value"][i] for i in num_indices]
        expected_num_sum = sum(num_values)
        actual_num_sums = [result_dict["num_sum"][i] for i in num_indices]
        assert all(
            sum == expected_num_sum for sum in actual_num_sums
        ), f"Incorrect sum for number {num}: {actual_num_sums} != {expected_num_sum}"

    for letter in ["A", "B", "C"]:
        for num in ["1", "2", "3"]:
            combined_indices = [
                i
                for i, (ltr, n) in enumerate(zip(result_dict["letter"], result_dict["num"]))
                if ltr == letter and n == num
            ]
            combined_values = [result_dict["value"][i] for i in combined_indices]
            expected_combined_sum = sum(combined_values)
            actual_combined_sums = [result_dict["combined_sum"][i] for i in combined_indices]
            assert all(
                sum == expected_combined_sum for sum in actual_combined_sums
            ), f"Incorrect sum for combination {letter}{num}: {actual_combined_sums} != {expected_combined_sum}"


def test_multi_window_agg_functions(make_df):
    """Test multiple window aggregation functions with different partition keys.

    Tests window functions with two different partition specifications:
    1. Partitioning by both category and group
    2. Partitioning by just category

    Using sum(), mean(), min(), and max() aggregations.
    """
    data = [
        {"category": "A", "group": 1, "value": 10},
        {"category": "A", "group": 1, "value": 20},
        {"category": "A", "group": 2, "value": 30},
        {"category": "A", "group": 2, "value": 40},
        {"category": "B", "group": 1, "value": 50},
        {"category": "B", "group": 1, "value": 60},
        {"category": "B", "group": 2, "value": 70},
        {"category": "B", "group": 2, "value": 80},
    ]

    df = make_df(data)

    multi_partition_window = Window().partition_by(["category", "group"])
    single_partition_window = Window().partition_by("category")

    result = df.select(
        col("category"),
        col("group"),
        col("value"),
        col("value").sum().over(multi_partition_window).alias("sum_multi"),
        col("value").mean().over(multi_partition_window).alias("avg_multi"),
        col("value").min().over(single_partition_window).alias("min_single"),
        col("value").max().over(single_partition_window).alias("max_single"),
    ).collect()

    result_dict = result.to_pydict()

    for category in ["A", "B"]:
        for group in [1, 2]:
            indices = [
                i
                for i, (cat, grp) in enumerate(zip(result_dict["category"], result_dict["group"]))
                if cat == category and grp == group
            ]

            values = [result_dict["value"][i] for i in indices]

            expected_sum = sum(values)
            expected_avg = sum(values) / len(values)

            for idx in indices:
                assert (
                    result_dict["sum_multi"][idx] == expected_sum
                ), f"Incorrect sum for {category}/{group}: {result_dict['sum_multi'][idx]} != {expected_sum}"
                assert (
                    abs(result_dict["avg_multi"][idx] - expected_avg) < 1e-10
                ), f"Incorrect avg for {category}/{group}: {result_dict['avg_multi'][idx]} != {expected_avg}"

    for category in ["A", "B"]:
        indices = [i for i, cat in enumerate(result_dict["category"]) if cat == category]

        values = [result_dict["value"][i] for i in indices]

        expected_min = min(values)
        expected_max = max(values)

        for idx in indices:
            assert (
                result_dict["min_single"][idx] == expected_min
            ), f"Incorrect min for {category}: {result_dict['min_single'][idx]} != {expected_min}"
            assert (
                result_dict["max_single"][idx] == expected_max
            ), f"Incorrect max for {category}: {result_dict['max_single'][idx]} != {expected_max}"


def test_without_source_columns_and_with_duplicate_window_functions(make_df):
    data = {
        "product": [],
        "category": [],
        "store": [],
        "revenue": [],
    }

    for store in [f"Store_{i}" for i in range(1, 3)]:
        idx = 1
        for category in ["Electronics", "Clothing", "Books"]:
            for _ in range(1, 3):
                data["product"].append(f"Product_{idx}")
                data["category"].append(category)
                data["store"].append(store)
                data["revenue"].append(random.randint(10, 100))
                idx += 1

    df = make_df(data)

    window_store = Window().partition_by("store")
    window_product = Window().partition_by("product")
    window_store_category = Window().partition_by(["store", "category"])

    result = df.select(
        col("revenue").sum().over(window_store).alias("store revenue sum"),
        (col("revenue") / col("revenue").sum().over(window_store)).alias("store revenue share"),
        col("revenue").sum().over(window_product).alias("product revenue sum"),
        (col("revenue") / col("revenue").sum().over(window_product)).alias("product revenue share"),
        col("revenue").sum().over(window_store_category).alias("store category revenue sum"),
        (col("revenue") - col("revenue").max().over(window_store_category)).alias("store category max revenue diff"),
    ).collect()

    pdf = pd.DataFrame(data)
    store_revenue_sum = pdf.groupby("store")["revenue"].transform("sum")
    store_revenue_share = pdf["revenue"] / store_revenue_sum
    product_revenue_sum = pdf.groupby("product")["revenue"].transform("sum")
    product_revenue_share = pdf["revenue"] / product_revenue_sum
    store_category_revenue_sum = pdf.groupby(["store", "category"])["revenue"].transform("sum")
    store_category_max_revenue = pdf.groupby(["store", "category"])["revenue"].transform("max")
    store_category_max_revenue_diff = pdf["revenue"] - store_category_max_revenue
    expected = pd.DataFrame(
        {
            "store revenue sum": store_revenue_sum,
            "store revenue share": store_revenue_share,
            "product revenue sum": product_revenue_sum,
            "product revenue share": product_revenue_share,
            "store category revenue sum": store_category_revenue_sum,
            "store category max revenue diff": store_category_max_revenue_diff,
        }
    )

    assert_df_equals(
        result.to_pandas(),
        expected,
        sort_key=[
            "store revenue sum",
            "product revenue sum",
            "store category revenue sum",
            "store category max revenue diff",
        ],
    )



from __future__ import annotations

import itertools

import numpy as np
import pyarrow as pa
import pytest

from daft import DataType, Series
from daft.utils import pyarrow_supports_fixed_shape_tensor
from tests.conftest import UuidType, get_tests_daft_runner_name
from tests.series import ARROW_FLOAT_TYPES, ARROW_INT_TYPES, ARROW_STRING_TYPES

ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())


class MockObject:
    def __init__(self, test_val):
        self.test_val = test_val


@pytest.mark.parametrize(
    "dtype, chunks", itertools.product(ARROW_FLOAT_TYPES + ARROW_INT_TYPES + ARROW_STRING_TYPES, [1, 2, 3, 10])
)
def test_series_concat(dtype, chunks) -> None:
    series = []
    for i in range(chunks):
        series.append(Series.from_pylist([i * j for j in range(i)]).cast(dtype=DataType.from_arrow_type(dtype)))

    concated = Series.concat(series)

    assert concated.datatype() == DataType.from_arrow_type(dtype)
    concated_list = concated.to_pylist()

    counter = 0
    for i in range(chunks):
        for j in range(i):
            val = i * j
            assert float(concated_list[counter]) == val
            counter += 1


@pytest.mark.parametrize(
    "dtype, chunks", itertools.product(ARROW_FLOAT_TYPES + ARROW_INT_TYPES + ARROW_STRING_TYPES, [1, 2, 3, 10])
)
def test_series_concat_with_slicing(dtype, chunks) -> None:
    series = []
    for i in range(chunks):
        s = Series.from_pylist([i] * 4).cast(dtype=DataType.from_arrow_type(dtype))
        series.append(s.slice(0, 2))

    concated = Series.concat(series)

    assert concated.datatype() == DataType.from_arrow_type(dtype)
    concated_list = concated.to_pylist()

    counter = 0
    for i in range(chunks):
        for _ in range(2):
            assert float(concated_list[counter]) == i
            counter += 1


@pytest.mark.parametrize("fixed", [False, True])
@pytest.mark.parametrize("chunks", [1, 2, 3, 10])
def test_series_concat_list_array(chunks, fixed) -> None:
    series = []
    arrow_type = pa.list_(pa.int64(), list_size=2 if fixed else -1)
    for i in range(chunks):
        series.append(Series.from_arrow(pa.array([[i + j, i * j] for j in range(i)], type=arrow_type)))

    concated = Series.concat(series)

    if fixed:
        assert concated.datatype() == DataType.fixed_size_list(DataType.int64(), 2)
    else:
        assert concated.datatype() == DataType.list(DataType.int64())
    concated_list = concated.to_pylist()

    counter = 0
    for i in range(chunks):
        for j in range(i):
            assert concated_list[counter][0] == i + j
            assert concated_list[counter][1] == i * j
            counter += 1


@pytest.mark.parametrize("chunks", [1, 2, 3, 10])
def test_series_concat_map_array(chunks) -> None:
    series = []
    for i in range(chunks):
        series.append(
            Series.from_arrow(
                pa.array(
                    [[("a", i + j), ("b", float(i * j))] for j in range(i)],
                    type=pa.map_(pa.string(), pa.float64()),
                )
            )
        )

    concated = Series.concat(series)

    assert concated.datatype() == DataType.map(DataType.string(), DataType.float64())
    concated_list = concated.to_pylist()
    counter = 0
    for i in range(chunks):
        for j in range(i):
            assert concated_list[counter][0][1] == i + j
            assert concated_list[counter][1][1] == float(i * j)
            counter += 1


@pytest.mark.parametrize("chunks", [1, 2, 3, 10])
def test_series_concat_struct_array(chunks) -> None:
    series = []
    for i in range(chunks):
        series.append(
            Series.from_arrow(
                pa.array(
                    [{"a": i + j, "b": float(i * j)} for j in range(i)],
                    type=pa.struct({"a": pa.int64(), "b": pa.float64()}),
                )
            )
        )

    concated = Series.concat(series)

    assert concated.datatype() == DataType.struct({"a": DataType.int64(), "b": DataType.float64()})
    concated_list = concated.to_pylist()

    counter = 0
    for i in range(chunks):
        for j in range(i):
            assert concated_list[counter]["a"] == i + j
            assert concated_list[counter]["b"] == float(i * j)
            counter += 1


@pytest.mark.skipif(
    not pyarrow_supports_fixed_shape_tensor(),
    reason=f"Arrow version {ARROW_VERSION} doesn't support the canonical tensor extension type.",
)
@pytest.mark.parametrize("chunks", [1, 2, 3, 10])
def test_series_concat_tensor_array_canonical(chunks) -> None:
    element_shape = (2, 2)
    num_elements_per_tensor = np.prod(element_shape)
    chunk_size = 3
    chunk_shape = (chunk_size,) + element_shape
    chunks = [
        np.arange(i * chunk_size * num_elements_per_tensor, (i + 1) * chunk_size * num_elements_per_tensor).reshape(
            chunk_shape
        )
        for i in range(chunks)
    ]
    ext_arrays = [pa.FixedShapeTensorArray.from_numpy_ndarray(chunk) for chunk in chunks]
    series = [Series.from_arrow(ext_array) for ext_array in ext_arrays]

    concated = Series.concat(series)

    assert concated.datatype() == DataType.tensor(
        DataType.from_arrow_type(ext_arrays[0].type.storage_type.value_type), (2, 2)
    )
    expected = [chunk[i] for chunk in chunks for i in range(len(chunk))]
    concated_arrow = concated.to_arrow()
    assert isinstance(concated_arrow.type, pa.FixedShapeTensorType)
    np.testing.assert_equal(concated_arrow.to_numpy_ndarray(), expected)


@pytest.mark.skipif(
    get_tests_daft_runner_name() == "ray",
    reason="pyarrow extension types aren't supported on Ray clusters.",
)
@pytest.mark.parametrize("chunks", [1, 2, 3, 10])
def test_series_concat_extension_type(uuid_ext_type, chunks) -> None:
    chunk_size = 3
    storage_arrays = [
        pa.array([f"{i}".encode() for i in range(j * chunk_size, (j + 1) * chunk_size)]) for j in range(chunks)
    ]
    ext_arrays = [pa.ExtensionArray.from_storage(uuid_ext_type, storage) for storage in storage_arrays]
    series = [Series.from_arrow(ext_array) for ext_array in ext_arrays]

    concated = Series.concat(series)

    assert concated.datatype() == DataType.extension(
        uuid_ext_type.NAME, DataType.from_arrow_type(uuid_ext_type.storage_type), ""
    )
    concated_arrow = concated.to_arrow()
    assert isinstance(concated_arrow.type, UuidType)
    assert concated_arrow.type == uuid_ext_type

    expected = uuid_ext_type.wrap_array(pa.concat_arrays(storage_arrays))

    assert concated_arrow == expected


@pytest.mark.parametrize("chunks", [1, 2, 3, 10])
def test_series_concat_pyobj(chunks) -> None:
    series = []
    for i in range(chunks):
        series.append(Series.from_pylist([MockObject(i * j) for j in range(i)], pyobj="force"))

    concated = Series.concat(series)

    assert concated.datatype() == DataType.python()
    concated_list = concated.to_pylist()

    counter = 0
    for i in range(chunks):
        for j in range(i):
            val = i * j
            assert concated_list[counter].test_val == val
            counter += 1


def test_series_concat_bad_input() -> None:
    mix_types_series = [Series.from_pylist([1, 2, 3]), []]
    with pytest.raises(TypeError, match="Expected a Series for concat"):
        Series.concat(mix_types_series)

    with pytest.raises(ValueError, match="Need at least 1 series"):
        Series.concat([])


def test_series_concat_dtype_mismatch() -> None:
    mix_types_series = [Series.from_pylist([1, 2, 3]), Series.from_pylist([1.0, 2.0, 3.0])]

    with pytest.raises(ValueError, match="concat requires all data types to match"):
        Series.concat(mix_types_series)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from daft.dependencies import pc

R = TypeVar("R")


if TYPE_CHECKING:
    from daft.expressions import Expression
    from daft.logical.schema import DataType


class ExpressionVisitor(ABC, Generic[R]):
    """ExpressionVisitor is an abstract base class for implementing the visitor pattern on expressions.

    Examples:
        >>> from daft.expressions import ExpressionVisitor, col, lit
        >>>
        >>> class PrintVisitor(ExpressionVisitor[None]):
        ...     def visit_col(self, name: str) -> None:
        ...         print(f"Column: {name}")
        ...
        ...     def visit_lit(self, value: Any) -> None:
        ...         print(f"Literal: {value}")
        ...
        ...     def visit_alias(self, expr: Expression, alias: str) -> None:
        ...         print(f"Alias: {alias}")
        ...         self.visit(expr)
        ...
        ...     def visit_cast(self, expr: Expression, dtype: DataType) -> None:
        ...         print(f"Cast: {dtype}")
        ...         self.visit(expr)
        ...
        ...     def visit_function(self, name: str, args: list[Expression]) -> None:
        ...         print(f"Function: {name}")
        ...         for arg in args:
        ...             self.visit(arg)
        >>> # Create an expression
        >>> expr = col("x").cast("int64").alias("y")
        >>>
        >>> # Visit the expression
        >>> visitor = PrintVisitor()
        >>> visitor.visit(expr)
        Alias: y
        Cast: Int64
        Column: x
    """

    def visit(self, expr: Expression) -> R:
        """Visit an arbitrary expression, invoking this visitor."""
        return expr._expr.accept(self)

    def visit_(self, name: str, args: list[Expression]) -> R:
        """Visit an arbitrary expression function, dispatching to an override if one exists."""
        if override := getattr(self, f"visit_{name}", None):
            return override(*args)
        else:
            return self.visit_function(name, args)

    @abstractmethod
    def visit_col(self, name: str) -> R:
        """Visit a col expression."""
        ...

    @abstractmethod
    def visit_lit(self, value: Any) -> R:
        """Visit a lit expression."""
        ...

    @abstractmethod
    def visit_alias(self, expr: Expression, alias: str) -> R:
        """Visit an alias expression."""
        ...

    @abstractmethod
    def visit_cast(self, expr: Expression, dtype: DataType) -> R:
        """Visit a cast expression."""
        ...

    @abstractmethod
    def visit_function(self, name: str, args: list[Expression]) -> R:
        """Visit a function call expression."""
        ...


class PredicateVisitor(ExpressionVisitor[R]):
    """PredicateVisitor is an ExpressionVisitor with helper methods for predicates."""

    @abstractmethod
    def visit_and(self, left: Expression, right: Expression) -> R:
        """Visit an and expression."""
        ...

    @abstractmethod
    def visit_or(self, left: Expression, right: Expression) -> R:
        """Visit an or expression."""
        ...

    @abstractmethod
    def visit_not(self, expr: Expression) -> R:
        """Visit a not expression."""
        ...

    @abstractmethod
    def visit_equal(self, left: Expression, right: Expression) -> R:
        """Visit an equals comparison predicate."""
        ...

    @abstractmethod
    def visit_not_equal(self, left: Expression, right: Expression) -> R:
        """Visit a not equals comparison predicate."""
        ...

    @abstractmethod
    def visit_less_than(self, left: Expression, right: Expression) -> R:
        """Visit a less than comparison predicate."""
        ...

    @abstractmethod
    def visit_less_than_or_equal(self, left: Expression, right: Expression) -> R:
        """Visit a less than or equal comparison predicate."""
        ...

    @abstractmethod
    def visit_greater_than(self, left: Expression, right: Expression) -> R:
        """Visit a greater than comparison predicate."""
        ...

    @abstractmethod
    def visit_greater_than_or_equal(self, left: Expression, right: Expression) -> R:
        """Visit a greater than or equal comparison predicate."""
        ...

    @abstractmethod
    def visit_between(self, expr: Expression, lower: Expression, upper: Expression) -> R:
        """Visit a between predicate."""
        ...

    @abstractmethod
    def visit_is_in(self, expr: Expression, items: list[Expression]) -> R:
        """Visit an is_in predicate."""
        ...

    @abstractmethod
    def visit_is_null(self, expr: Expression) -> R:
        """Visit an is_null predicate."""
        ...

    @abstractmethod
    def visit_not_null(self, expr: Expression) -> R:
        """Visit an not_null predicate."""
        ...


class _PyArrowExpressionVisitor(PredicateVisitor[pc.Expression]):
    """This visitor does a tree fold into the pyarrow.compute.Expression domain."""

    def visit_col(self, name: str) -> pc.Expression:
        """Convert the daft column to pc field reference by name."""
        return pc.field(name)

    def visit_lit(self, value: Any) -> pc.Expression:
        """Convert the Literal to a pyarrow.compute.Scalar without check the type.

        From the pyarrow scalar docs:
        > value : bool, int, float or string
        > Python value of the scalar. Note that only a subset of types are currently supported.

        Coincidentally, this is what term is currently limited to.
        """
        return pc.scalar(value)

    def visit_alias(self, expr: Expression, alias: str) -> pc.Expression:
        """Convert an alias 'expression' by ... ignoring it .. as these aren't supposed to be expressions."""
        return self.visit(expr)

    def visit_cast(self, expr: Expression, dtype: DataType) -> pc.Expression:
        """Converts the cast with default safety and cast options because daft does not have these options."""
        pc_expr = self.visit(expr)
        pc_type = dtype.to_arrow_dtype()
        return pc_expr.cast(pc_type)

    def visit_list(self, items: list[Expression]) -> pc.Expression:
        raise ValueError("pyarrow.compute does not have a make_list function.")

    def visit_and(self, left: Expression, right: Expression) -> pc.Expression:
        pc_lhs = self.visit(left)
        pc_rhs = self.visit(right)
        return pc_lhs & pc_rhs

    def visit_or(self, left: Expression, right: Expression) -> pc.Expression:
        pc_lhs = self.visit(left)
        pc_rhs = self.visit(right)
        return pc_lhs | pc_rhs

    def visit_not(self, expr: Expression) -> pc.Expression:
        pc_expr = self.visit(expr)
        return ~pc_expr

    def visit_equal(self, left: Expression, right: Expression) -> pc.Expression:
        pc_lhs = self.visit(left)
        pc_rhs = self.visit(right)
        return pc_lhs == pc_rhs

    def visit_not_equal(self, left: Expression, right: Expression) -> pc.Expression:
        pc_lhs = self.visit(left)
        pc_rhs = self.visit(right)
        return pc_lhs != pc_rhs

    def visit_less_than(self, left: Expression, right: Expression) -> pc.Expression:
        pc_lhs = self.visit(left)
        pc_rhs = self.visit(right)
        return pc_lhs < pc_rhs

    def visit_less_than_or_equal(self, left: Expression, right: Expression) -> pc.Expression:
        pc_lhs = self.visit(left)
        pc_rhs = self.visit(right)
        return pc_lhs <= pc_rhs

    def visit_greater_than(self, left: Expression, right: Expression) -> pc.Expression:
        pc_lhs = self.visit(left)
        pc_rhs = self.visit(right)
        return pc_lhs > pc_rhs

    def visit_greater_than_or_equal(self, left: Expression, right: Expression) -> pc.Expression:
        pc_lhs = self.visit(left)
        pc_rhs = self.visit(right)
        return pc_lhs >= pc_rhs

    def visit_between(self, expr: Expression, lower: Expression, upper: Expression) -> pc.Expression:
        """Convert between using le and ge since the bounds are inclusive."""
        pc_expr = self.visit(expr)
        pc_lower = self.visit(lower)
        pc_upper = self.visit(upper)
        return (pc_lower <= pc_expr) & (pc_expr <= pc_upper)

    def visit_is_in(self, expr: Expression, items: list[Expression]) -> pc.Expression:
        pc_expr = self.visit(expr)
        pc_items = [item.as_py() for item in items]
        return pc_expr.isin(pc_items)

    def visit_is_null(self, expr: Expression) -> pc.Expression:
        pc_expr = self.visit(expr)
        return pc_expr.is_null()

    def visit_not_null(self, expr: Expression) -> pc.Expression:
        pc_expr = self.visit(expr)
        return ~pc_expr.is_null()

    def visit_function(self, name: str, args: list[Expression]) -> pc.Expression:
        """Converting using either the EXACT function name or an override, otherwise error."""
        if hasattr(self, "_" + name):
            # special form, call the overriding method
            return getattr(self, "_" + name)(*args)
        else:
            # normal form, call the pyarrow.compute method
            pc_func = self._get_pc_func(name)
            pc_args = [self.visit(arg) for arg in args]
            return pc_func(*pc_args)

    def _get_pc_func(self, name: str) -> Any:
        """Resolve the pyarrow.compute function from the module, otherwise error."""
        try:
            pc_name = self._PC_FUNCTION_OVERRIDES.get(name, name)
            pc_func = getattr(pc, pc_name)
            return pc_func
        except AttributeError:
            raise ValueError(
                f"pyarrow.compute has no function '{name}', please see: https://arrow.apache.org/docs/python/api/compute.html."
            )

    # FUNCTION OVERRIDES
    #
    #   This maps daft functions to pyarrow.compute expression functions.
    #   https://arrow.apache.org/docs/python/api/compute.html
    #
    #   Normal forms map to strings.
    #   Special forms map to the special translation functions.
    #
    _PC_FUNCTION_OVERRIDES: dict[str, str] = {
        "plus": "add",
        "minus": "subtract",
        # math overrides
        "log": "logb",
        "negative": "negate",
        "arcsin": "asin",
        "arccos": "acos",
        "arctan": "atan",
        "arctan2": "atan2",
        # string overrides
        "capitalize": "utf8_capitalize",
    }

    def _round(self, input: Expression, precision: Expression) -> pc.Expression:
        pc_input = self.visit(input)
        pc_precision = precision.as_py()  # must be a literal
        return pc.round(pc_input, pc_precision)

    def _count_matches(
        self, input: Expression, pattern: Expression, whole_words: Expression, case_sensitive: Expression
    ) -> pc.Expression:
        pc_strings = self.visit(input)
        pc_pattern = pattern.as_py()  # must be literal
        pc_ignore_case = not case_sensitive.as_py()
        return pc.count_substring(pc_strings, pc_pattern, ignore_case=pc_ignore_case)

    def _contains(self, input: Expression, substring: Expression) -> pc.Expression:
        pc_strings = self.visit(input)
        pc_pattern = substring.as_py()  # must be literal
        return pc.match_substring(pc_strings, pc_pattern)

    def _ends_with(self, input: Expression, suffix: Expression) -> pc.Expression:
        pc_strings = self.visit(input)
        pc_pattern = suffix.as_py()  # must be literal
        return pc.ends_with(pc_strings, pc_pattern)

    def _starts_with(self, input: Expression, prefix: Expression) -> pc.Expression:
        pc_strings = self.visit(input)
        pc_pattern = prefix.as_py()  # must be literal
        return pc.starts_with(pc_strings, pc_pattern)



from __future__ import annotations

import itertools

import pyarrow as pa
import pytest

from daft import col
from daft.logical.schema import Schema
from daft.recordbatch import MicroPartition
from daft.series import Series
from tests.recordbatch import daft_int_types, daft_numeric_types


@pytest.mark.parametrize(
    "mp",
    [
        MicroPartition.from_pydict({"a": pa.array([], type=pa.int64())}),  # 1 empty table
        MicroPartition.empty(Schema.from_pyarrow_schema(pa.schema({"a": pa.int64()}))),  # No tables
    ],
)
def test_micropartitions_take_empty(mp) -> None:
    assert len(mp) == 0

    indices = Series.from_arrow(pa.array([], type=pa.int64()))
    taken = mp.take(indices)
    assert len(taken) == 0
    assert taken.column_names() == ["a"]
    assert taken.to_pydict() == {"a": []}

    indices = Series.from_arrow(pa.array([1], type=pa.int64()))

    with pytest.raises(BaseException, match="index out of bounds"):
        taken = mp.take(indices)


@pytest.mark.parametrize(
    "mp",
    [
        MicroPartition.from_pydict({"a": [1, 2, 3, 4]}),  # 1 table
        MicroPartition.concat(
            [
                MicroPartition.from_pydict({"a": pa.array([], type=pa.int64())}),
                MicroPartition.from_pydict({"a": [1]}),
                MicroPartition.from_pydict({"a": [2, 3, 4]}),
            ]
        ),  # 3 tables
    ],
)
def test_micropartitions_take(mp: MicroPartition) -> None:
    assert mp.column_names() == ["a"]
    assert len(mp) == 4

    indices = Series.from_pylist([0, 1])
    taken = mp.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["a"]
    assert taken.to_pydict() == {"a": [1, 2]}

    indices = Series.from_pylist([3, 2])
    taken = mp.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["a"]
    assert taken.to_pydict() == {"a": [4, 3]}

    indices = Series.from_pylist([3, 2, 2, 2, 3])
    taken = mp.take(indices)
    assert len(taken) == 5
    assert taken.column_names() == ["a"]
    assert taken.to_pydict() == {"a": [4, 3, 3, 3, 4]}


@pytest.mark.parametrize("data_dtype, idx_dtype", itertools.product(daft_numeric_types, daft_int_types))
def test_table_take_numeric(data_dtype, idx_dtype) -> None:
    pa_table = pa.Table.from_pydict({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("a").cast(data_dtype), col("b")])

    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    indices = Series.from_pylist([0, 1]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": [1, 2], "b": [5, 6]}

    indices = Series.from_pylist([3, 2]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": [4, 3], "b": [8, 7]}

    indices = Series.from_pylist([3, 2, 2, 2, 3]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 5
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": [4, 3, 3, 3, 4], "b": [8, 7, 7, 7, 8]}


@pytest.mark.parametrize("idx_dtype", daft_int_types)
def test_table_take_str(idx_dtype) -> None:
    pa_table = pa.Table.from_pydict({"a": ["1", "2", "3", "4"], "b": ["5", "6", "7", "8"]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    indices = Series.from_pylist([0, 1]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": ["1", "2"], "b": ["5", "6"]}

    indices = Series.from_pylist([3, 2]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": ["4", "3"], "b": ["8", "7"]}

    indices = Series.from_pylist([3, 2, 2, 2, 3]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 5
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": ["4", "3", "3", "3", "4"], "b": ["8", "7", "7", "7", "8"]}


@pytest.mark.parametrize("idx_dtype", daft_int_types)
def test_table_take_bool(idx_dtype) -> None:
    pa_table = pa.Table.from_pydict({"a": [False, True, False, True], "b": [True, False, True, False]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    indices = Series.from_pylist([0, 1]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": [False, True], "b": [True, False]}

    indices = Series.from_pylist([3, 2]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": [True, False], "b": [False, True]}

    indices = Series.from_pylist([3, 2, 2, 2, 3]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 5
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": [True, False, False, False, True], "b": [False, True, True, True, False]}


@pytest.mark.parametrize("idx_dtype", daft_int_types)
def test_table_take_null(idx_dtype) -> None:
    pa_table = pa.Table.from_pydict({"a": [None, None, None, None], "b": [None, None, None, None]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    indices = Series.from_pylist([0, 1]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": [None, None], "b": [None, None]}


def test_table_take_pyobject() -> None:
    objects = [object(), None, object(), object()]
    daft_recordbatch = MicroPartition.from_pydict({"objs": objects})
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["objs"]

    indices = Series.from_pylist([0, 1])

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["objs"]

    assert taken.to_pydict()["objs"] == objects[:2]

    indices = Series.from_pylist([3, 2])

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["objs"]

    assert taken.to_pydict()["objs"] == [objects[3], objects[2]]

    indices = Series.from_pylist([3, 2, 2, 2, 3])

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 5
    assert taken.column_names() == ["objs"]

    assert taken.to_pydict()["objs"] == [objects[3], objects[2], objects[2], objects[2], objects[3]]


@pytest.mark.parametrize("idx_dtype", daft_int_types)
def test_table_take_fixed_size_list(idx_dtype) -> None:
    pa_table = pa.Table.from_pydict(
        {
            "a": pa.array([[1, 2], [3, None], None, [None, None]], type=pa.list_(pa.int64(), 2)),
            "b": pa.array([[4, 5], [6, None], None, [None, None]], type=pa.list_(pa.int64(), 2)),
        }
    )
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    indices = Series.from_pylist([0, 1]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": [[1, 2], [3, None]], "b": [[4, 5], [6, None]]}

    indices = Series.from_pylist([3, 2]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 2
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {"a": [[None, None], None], "b": [[None, None], None]}

    indices = Series.from_pylist([3, 2, 2, 2, 3]).cast(idx_dtype)

    taken = daft_recordbatch.take(indices)
    assert len(taken) == 5
    assert taken.column_names() == ["a", "b"]

    assert taken.to_pydict() == {
        "a": [[None, None], None, None, None, [None, None]],
        "b": [[None, None], None, None, None, [None, None]],
    }



from __future__ import annotations

import pytest


@pytest.mark.parametrize("series", ["Expression", "SQL"], indirect=True)
def test_jq(series):
    items = [
        '{ "a": null, "b": true, "c": 1, "d": 1.1, "e": "ABC" }',
        '{ "a": null, "b": true, "c": 2, "d": 2.2, "e": "DEF" }',
        '{ "a": null, "b": false, "c": 3, "d": 3.3, "e": "GHI" }',
        '{ "a": null, "b": false, "c": 3, "d": 4.4, "e": "JKL" }',
        '{ "a": null, "b": null, "c": null, "d": null, "e": null }',
    ]
    # sanity selector tests for various types
    assert series(items).jq(".a") == ["null", "null", "null", "null", "null"]
    assert series(items).jq(".b") == ["true", "true", "false", "false", "null"]
    assert series(items).jq(".c") == ["1", "2", "3", "3", "null"]
    assert series(items).jq(".d") == ["1.1", "2.2", "3.3", "4.4", "null"]
    assert series(items).jq(".e") == ['"ABC"', '"DEF"', '"GHI"', '"JKL"', "null"]


@pytest.mark.parametrize("series", ["Expression", "SQL"], indirect=True)
def test_jq_with_nulls(series):
    items = [
        None,
        '{ "a": null, "b": true, "c": 1, "d": 1.1, "e": "ABC" }',
        None,
        None,
        '{ "a": null, "b": false, "c": 2, "d": 2.2, "e": "DEF" }',
        None,
    ]
    # sanity selector tests for various types
    assert series(items).jq(".a") == [None, "null", None, None, "null", None]
    assert series(items).jq(".b") == [None, "true", None, None, "false", None]
    assert series(items).jq(".c") == [None, "1", None, None, "2", None]
    assert series(items).jq(".d") == [None, "1.1", None, None, "2.2", None]
    assert series(items).jq(".e") == [None, '"ABC"', None, None, '"DEF"', None]


@pytest.mark.parametrize("series", ["Expression", "SQL"], indirect=True)
def test_jq_simple_nested_path(series):
    items = [
        '{ "a": { "b": { "c": "deep" } } }',
        '{ "a": { "b": { "c": "shallow" } } }',
        '{ "a": { "b": { "c": "medium" } } }',
        '{ "a": { "b": { "c": "high" } } }',
    ]
    assert series(items).jq(".a.b.c") == ['"deep"', '"shallow"', '"medium"', '"high"']


@pytest.mark.parametrize("series", ["Expression", "SQL"], indirect=True)
def test_jq_array_selection(series):
    items = [
        '{ "a": { "b": [1, 2, 3] } }',
        '{ "a": { "b": [4, 5, 6] } }',
        '{ "a": { "b": [7, 8, 9] } }',
        '{ "a": { "b": [10, 11, 12] } }',
    ]
    assert series(items).jq(".a.b[0]") == ["1", "4", "7", "10"]


@pytest.mark.parametrize("series", ["Expression", "SQL"], indirect=True)
def test_jq_nested_array_selection(series):
    items = [
        '{ "a": [{ "b": "nested1" }, { "b": "nested2" }] }',
        '{ "a": [{ "b": "nested3" }, { "b": "nested4" }] }',
        '{ "a": [{ "b": "nested5" }, { "b": "nested6" }] }',
        '{ "a": [{ "b": "nested7" }, { "b": "nested8" }] }',
    ]
    assert series(items).jq(".a[0].b") == ['"nested1"', '"nested3"', '"nested5"', '"nested7"']


@pytest.mark.parametrize("series", ["Expression", "SQL"], indirect=True)
def test_jq_nested_array_with_array_selection(series):
    items = [
        '{ "a": [{ "b": [1, 2] }, { "b": [3, 4] }] }',
        '{ "a": [{ "b": [5, 6] }, { "b": [7, 8] }] }',
        '{ "a": [{ "b": [9, 10] }, { "b": [11, 12] }] }',
        '{ "a": [{ "b": [13, 14] }, { "b": [15, 16] }] }',
    ]
    assert series(items).jq(".a[0].b[0]") == ["1", "5", "9", "13"]


@pytest.mark.parametrize("series", ["Expression", "SQL"], indirect=True)
def test_jq_array_wildcard(series):
    items = [
        '{ "a": [{ "b": "nested1" }, { "b": "nested2" }] }',
        '{ "a": [{ "b": "nested3" }, { "b": "nested4" }] }',
        '{ "a": [{ "b": "nested5" }, { "b": "nested6" }] }',
        '{ "a": [{ "b": "nested7" }, { "b": "nested8" }] }',
    ]
    assert series(items).jq(".a[].b") == [
        '["nested1","nested2"]',
        '["nested3","nested4"]',
        '["nested5","nested6"]',
        '["nested7","nested8"]',
    ]


@pytest.mark.parametrize("series", ["Expression", "SQL"], indirect=True)
def test_jq_permissive_paths(series):
    items = [
        '{ "a": 1 }',  # <-- !! MISSING !! so this will return a non-match "" rather than "null"
        '{ "a": { "b": null } }',
        '{ "a": { "b": true } }',
    ]
    assert series(items).jq(".a.b?") == ["", "null", "true"]


@pytest.mark.parametrize("series", ["Expression", "SQL"], indirect=True)
def test_jq_permissive_types(series):
    items = [
        '{ "a": true }',  # <-- type exception, so we'll get a non-match ""
        '{ "a": [ 1 ] }',  # <-- out-of-bounds, so we'll get a "null"
        '{ "a": [ 1, 2 ] }',
    ]
    assert series(items).jq("(.a | arrays?)[1]") == ["", "null", "2"]


@pytest.mark.parametrize("series", ["Expression", "SQL"], indirect=True)
def test_jq_complex_transformation(series):
    items = [
        """{
            "users": [
                {
                    "name": "Alice",
                    "scores": [85, 92, 78]
                },
                {
                    "name": "Bob",
                    "scores": [91, 88, 95]
                }
            ]
        }""",
        """{
            "users": [
                {
                    "name": "Charlie",
                    "scores": [82, 89, 94]
                },
                {
                    "name": "Dave",
                    "scores": [88, 91, 86]
                }
            ]
        }""",
    ]
    # some jq filter to get average scores because jq is not jsonpath
    assert series(items).jq(".users[] | {name: .name, avg_score: ([.scores[]] | add / length | round)}") == [
        '[{"name":"Alice","avg_score":85},{"name":"Bob","avg_score":91}]',
        '[{"name":"Charlie","avg_score":88},{"name":"Dave","avg_score":88}]',
    ]


def test_jq_with_deserialize():
    import daft
    from daft import DataType as dt
    from daft import col

    # here's our raw sample data which is just some json dump from a sensor
    df = daft.from_pydict(
        {
            "sample": [
                '{ "x": 1 }',  # missing y, we'll insert 0 in its place
                '{ "x": 1, "y": 1 }',  # ok
                '"HELLO, WORLD!"',  # you're not supposed to be here..
                '{ "x": 3, "y": 3 }',  # ok
                '{ "x": 4, "y": 4 }',  # ok
                '{ "x": false }',  # wrong data type..
            ]
        }
    )

    # select all objects, using 0 as the default for missing keys
    filter = """
        (. | objects?) | { x: .x // 0, y: .y // 0 }
    """

    # our point type is an x/y pair.
    point_t = dt.struct({"x": dt.int64(), "y": dt.int64()})

    # we have the successfully extracted each sample point, now deserialize into our type.
    points = (df.select(col("sample").jq(filter).try_deserialize("json", point_t).alias("point"))).drop_null()

    # now find the max from the origin, no need to sqrt it.
    p = col("point")
    furthest_point = (
        points.with_column("distance", p["x"] * p["x"] + p["y"] * p["y"])
        .sort("distance", desc=True)
        .limit(1)
        .select(p)
        .to_pydict()["point"][0]
    )

    assert furthest_point == {"x": 4, "y": 4}



from __future__ import annotations

import pytest

import daft
from daft import col


def test_aggs_sql():
    df = daft.from_pydict(
        {
            "id": [1, 2, 3, 3, 3, 3, 2, 1, 3, 1],
            "values": [1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5],
            "floats": [0.01, 0.011, 0.01047, 0.02, 0.019, 0.018, 0.017, 0.016, 0.015, 0.014],
        }
    )
    expected = (
        df.agg(
            [
                col("values").sum().alias("sum"),
                col("values").mean().alias("mean"),
                col("values").min().alias("min"),
                col("values").max().alias("max"),
                col("values").count().alias("count"),
                col("values").count_distinct().alias("count_distinct"),
                col("values").stddev().alias("std"),
            ]
        )
        .collect()
        .to_pydict()
    )

    actual = (
        daft.sql("""
    SELECT
        sum(values) as sum,
        mean(values) as mean,
        min(values) as min,
        max(values) as max,
        count(values) as count,
        count(distinct values) as count_distinct,
        stddev(values) as std
    FROM df
    """)
        .collect()
        .to_pydict()
    )

    assert actual == expected


@pytest.mark.parametrize(
    "agg,cond,expected",
    [
        ("sum(values)", "sum(values) > 10", {"values": [22.0, 29.5]}),
        ("sum(values)", "values > 10", {"values": [22.0, 29.5]}),
        ("sum(values) as sum_v", "sum(values) > 10", {"sum_v": [22.0, 29.5]}),
        ("sum(values) as sum_v", "sum_v > 10", {"sum_v": [22.0, 29.5]}),
        ("count(*) as cnt", "cnt > 2", {"cnt": [4, 5]}),
        ("count(*) as cnt", "count(*) > 2", {"cnt": [4, 5]}),
        ("count(1) as cnt", "cnt > 2", {"cnt": [4, 5]}),
        ("count(1) as cnt", "count(1) > 2", {"cnt": [4, 5]}),
        ("count(*)", "count(*) > 2", {"count": [4, 5]}),
        ("count(1)", "count(1) > 2", {"count": [4, 5]}),
        ("count(*) as cnt", "sum(values) > 10", {"cnt": [4, 5]}),
        ("count(1) as cnt", "sum(values) > 10", {"cnt": [4, 5]}),
        ("count(true) as cnt", "sum(values) > 10", {"cnt": [4, 5]}),
        ("count(false) as cnt", "sum(values) > 10", {"cnt": [4, 5]}),
        ("count([1]) as cnt", "sum(values) > 10", {"cnt": [4, 5]}),
        ("count(null) as null", "sum(values) > 10", {"null": [0, 0]}),
        # duplicates of the above 4 `count` tests but for count-distinct
        ("count(distinct values) as count_distinct", "count_distinct > 2", {"count_distinct": [3, 5]}),
        ("count(distinct values) as count_distinct", "count(distinct values) > 2", {"count_distinct": [3, 5]}),
        ("count(distinct values)", "count(distinct values) > 2", {"values": [3, 5]}),
        ("count(distinct values) as count_distinct", "sum(values) > 10", {"count_distinct": [3, 5]}),
        ("sum(values), count(*)", "id > 1", {"values": [10.0, 29.5], "count": [2, 5]}),
    ],
)
def test_having(agg, cond, expected):
    df = daft.from_pydict(
        {
            "id": [1, 2, 3, 3, 3, 3, 2, 1, 3, 1, 1],
            "values": [1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 1.5],
        }
    )
    bindings = {"df": df}

    actual = daft.sql(
        f"""
    SELECT
    {agg},
    from df
    group by id
    having {cond}
    order by id
    """,
        **bindings,
    ).to_pydict()

    assert actual == expected


def test_having_non_grouped():
    df = daft.from_pydict(
        {
            "id": [1, 2, 3, 3, 3, 3, 2, 1, 3, 1],
            "values": [1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5],
            "floats": [0.01, 0.011, 0.01047, 0.02, 0.019, 0.018, 0.017, 0.016, 0.015, 0.014],
        }
    )
    bindings = {"df": df}

    actual = daft.sql(
        """
    SELECT
    count(*) ,
    from df
    having sum(values) > 40
    """,
        **bindings,
    ).to_pydict()

    assert actual == {"count": [10]}


def test_simple_rollup():
    data = {"dept": ["IT", "IT", "HR", "HR"], "year": [2022, 2023, 2022, 2023], "cost": [100, 200, 300, 400]}
    expected = {"dept": ["HR", "IT", None], "year": [None, None, None], "total": [700, 300, 1000]}
    df = daft.from_pydict(data)
    bindings = {"df": df}
    sql = """
    SELECT
        dept,
        year,
        SUM(cost) as total
    FROM df
    GROUP BY ROLLUP(dept, year)
    HAVING year IS NULL
    ORDER BY dept NULLS LAST
    """
    actual = daft.sql(sql, **bindings).to_pydict()
    assert actual == expected


def test_rollup_null_handling():
    data = {"dept": ["IT", "IT", "HR", "HR"], "year": [2022, 2023, 2022, 2023], "cost": [100, 200, 300, 400]}
    expected = {"dept": ["HR", "IT", None], "year": [None, None, None], "total": [700, 300, 1000]}
    df = daft.from_pydict(data)
    bindings = {"df": df}

    sql = """
    SELECT dept, year, SUM(cost) as total
    FROM df
    GROUP BY ROLLUP(dept, year)
    HAVING year IS NULL
    ORDER BY dept NULLS LAST
    """
    actual = daft.sql(sql, **bindings).to_pydict()
    assert actual == expected


def test_rollup_multiple_aggs():
    data = {
        "region": ["East", "East", "West", "West"],
        "product": ["A", "B", "A", "B"],
        "qty": [10, 20, 30, 40],
        "price": [5, 8, 6, 9],
    }
    expected = {"region": ["East", "West", None], "total_qty": [30, 70, 100], "avg_price": [6.5, 7.5, 7.0]}

    df = daft.from_pydict(data)
    bindings = {"df": df}

    sql = """
    SELECT
        region,
        SUM(qty) as total_qty,
        AVG(price) as avg_price
    FROM df
    GROUP BY ROLLUP(region, product)
    HAVING product IS NULL
    ORDER BY region NULLS LAST
    """
    actual = daft.sql(sql, **bindings).to_pydict()
    assert actual == expected


def test_groupby_rollup():
    data = {
        "region": ["EMEA", "EMEA", "APAC", "APAC", "AMER", "AMER"],
        "product": ["HW", "SW", "HW", "SW", "HW", "SW"],
        "channel": ["Direct", "Partner", "Direct", "Partner", "Direct", "Partner"],
        "sales": [100, 200, 300, 400, 500, 600],
        "costs": [50, 80, 120, 160, 200, 240],
    }
    expected = {
        "region": ["AMER", "All Regions"],
        "product": ["All Products", "All Products"],
        "channel": ["All Channels", "All Channels"],
        "total_sales": [1100, 2100],
        "total_costs": [440, 850],
        "profit": [660, 1250],
        "roi": [2.5, 2.4705882352941178],
    }
    df = daft.from_pydict(data)
    bindings = {"sales": df}

    sql = """
    SELECT
        COALESCE(region, 'All Regions') as region,
        COALESCE(product, 'All Products') as product,
        COALESCE(channel, 'All Channels') as channel,
        SUM(sales) as total_sales,
        SUM(costs) as total_costs,
        SUM(sales - costs) as profit,
        SUM(sales) / SUM(costs) as roi
    FROM df
    GROUP BY ROLLUP(region, product, channel)
    HAVING SUM(sales) > 1000
    ORDER BY region, product, channel;
    """
    actual = daft.sql(sql, **bindings).to_pydict()
    assert actual == expected


@pytest.mark.parametrize(
    "query",
    [
        "SELECT capitalize(strings) as s, count(*) from data group by s",
        "SELECT capitalize(strings) as s, count(*) from data group by capitalize(strings)",
        "SELECT strings, count(*) from data group by capitalize(strings)",
        "SELECT strings as s, count(*) from data group by capitalize(strings)",
    ],
)
def test_various_groupby_positions(query):
    data = daft.from_pydict({"strings": ["foo", "bar"]})
    res = daft.sql(query, **{"data": data})
    if "s" in res.column_names:
        res = res.sort("s").to_pydict()
        assert res == {"s": ["Bar", "Foo"], "count": [1, 1]}
    else:
        res = res.sort("strings").to_pydict()
        assert res == {"strings": ["Bar", "Foo"], "count": [1, 1]}


def test_group_by_with_lit_selection():
    df = daft.from_pydict({"ids": [1, 1, 2, 2, 3, 4], "values": [10, 20, 30, 40, 50, 60]})

    res = daft.sql("select 0 from df group by ids", df=df)
    expected = {"literal": [0, 0, 0, 0]}
    assert res.to_pydict() == expected
